"""
Objects that generate component assemblies for a model.
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator


# pylint: disable=dangerous-default-value

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Type
from typing import Union
from typing import Optional
from typing import Callable
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from .query import ElmQuery
from .node_gen import NodeGenerator
from ..ops.uniaxial_material import UniaxialMaterial
from ..physical_material import PhysicalMaterial
from ..mesh import Mesh
from ..ops.node import Node
from ..component_assembly import ComponentAssembly
from ..ops.element import TrussBar
from ..ops.element import ElasticBeamColumn
from ..ops.element import DispBeamColumn
from ..ops.element import ZeroLength
from ..ops.element import TwoNodeLink
from ..ops.element import GeomTransf
from ..ops.element import Lobatto
from ..ops.section import ElasticSection
from ..ops.section import FiberSection
from ..preprocessing.split_component import split_component
from ..transformations import local_axes_from_points_and_angle
from ..transformations import transformation_matrix
from ..gen import zerolength_gen
from .. import common


if TYPE_CHECKING:
    from ..model import Model
    from ..level import Level


nparr = npt.NDArray[np.float64]


def retrieve_snap_pt_global_offset(
    placement: str,
    section: ElasticSection | FiberSection,
    p_i: nparr,
    p_j: nparr,
    angle: float,
) -> nparr:
    """
    Returns the necessary offset to connect an element at a specified
    snap point of the section.

    Arguments:
      placement: Placement tag. Can be any of "centroid",
        "top_center", "top_left", "top_right", "center_left",
        "center_right", "bottom_center", "bottom_left", "bottom_right"
      section: Section object.
      p_i: Internal point at the i-end.
      p_j: Internal point at the j-end.

    Returns:
      The offset.

    """

    if section.snap_points and (placement != "centroid"):
        # obtain offset from section (local system)
        d_z, d_y = section.snap_points[placement]
        sec_offset_local: nparr = np.array([0.00, d_y, d_z])
        # retrieve local coordinate system
        x_axis, y_axis, z_axis = local_axes_from_points_and_angle(
            p_i, p_j, angle
        )
        t_glob_to_loc = transformation_matrix(x_axis, y_axis, z_axis)
        t_loc_to_glob = t_glob_to_loc.T
        sec_offset_global = t_loc_to_glob @ sec_offset_local
    else:
        sec_offset_global = np.zeros(3)
    return sec_offset_global


def beam_placement_lookup(
    x_coord,
    y_coord,
    query,
    ndg,
    lvls,
    key,
    user_offset,
    section_offset,
    split_existing,
    snap,
):
    """
    Performs lookup operations before placing a beam-functioning
    component assembly to determine how to connect it with
    respect to the other existing objects in the model.

    """

    lvl = lvls[key]
    node = query.search_node_lvl(x_coord, y_coord, lvl.uid)
    pinit = np.array((x_coord, y_coord, lvl.elevation)) + user_offset
    e_o = user_offset.copy() + section_offset
    if not node:
        if split_existing:
            node, offset = split_component(split_existing, pinit)
            e_o += offset
        else:
            node = ndg.add_node_lvl(x_coord, y_coord, key)
    else:
        # first check if a panel zone or other type of joint-like
        # component assembly exists at that node
        result_node = None
        components = query.retrieve_components_from_nodes([node], lvl.uid)
        for component in components.values():
            if component.component_purpose == "steel_W_panel_zone":
                if snap in [
                    "middle_front",
                    "middle_back",
                    "top_node",
                    "bottom_node",
                ]:
                    result_node = component.external_nodes.named_contents[snap]
                    e_o += np.array(
                        (0.00, 0.00, node.coords[2] - result_node.coords[2])
                    )
                    node = result_node
                    return node, e_o
                if snap in [
                    "centroid",
                    "top_center",
                    "top_left",
                    "top_right",
                    "center_left",
                    "center_right",
                    "bottom_center",
                    "bottom_left",
                    "bottom_right",
                ]:
                    elm = component.elements.named_contents["elm_interior"]
                    d_z, d_y = elm.section.snap_points[snap]
                    sec_offset_local: nparr = -np.array([0.00, d_y, d_z])
                    # retrieve local coordinate system
                    x_axis = elm.geomtransf.x_axis
                    y_axis = elm.geomtransf.y_axis
                    z_axis = elm.geomtransf.z_axis
                    t_glob_to_loc = transformation_matrix(
                        x_axis, y_axis, z_axis
                    )
                    t_loc_to_glob = t_glob_to_loc.T
                    sec_offset_global = t_loc_to_glob @ sec_offset_local
                    result_node = node
                    e_o += sec_offset_global
                    return node, e_o
                raise ValueError(f"Unsupported snap keyword: {snap}")

        # else check if a column-like component assembly exists
        if key - 1 in lvls:
            node_below = query.search_node_lvl(x_coord, y_coord, key - 1)
            if node_below:
                column = query.search_connectivity([node, node_below])
                if column:
                    elms = []
                    for dctkey in column.element_connectivity().keys():
                        if node.uid in dctkey:
                            elms.append(column.element_connectivity()[dctkey])
                    assert elms, "There should be an element here."
                    assert (
                        len(elms) == 1
                    ), "There should only be one element here."
                    elm = elms[0]
                    # obtain offset from section (local system)
                    if hasattr(elm, 'section'):
                        if elm.section.snap_points:
                            d_z, d_y = elm.section.snap_points[snap]
                            sec_offset_local = -np.array([0.00, d_y, d_z])
                            # retrieve local coordinate system
                            x_axis = elm.geomtransf.x_axis
                            y_axis = elm.geomtransf.y_axis
                            z_axis = elm.geomtransf.z_axis
                            t_glob_to_loc = transformation_matrix(
                                x_axis, y_axis, z_axis
                            )
                            t_loc_to_glob = t_glob_to_loc.T
                            sec_offset_global = (
                                t_loc_to_glob @ sec_offset_local
                            )
                            e_o += sec_offset_global
    return node, e_o


def look_for_panel_zone(node: Node, lvl: Level, query: ElmQuery) -> Node:
    """
    Determines if a panel zone joint component assembly is present
    at the specified node.

    """

    components = query.retrieve_components_from_nodes([node], lvl.uid)
    result_node = node
    for component in components.values():
        if component.component_purpose == "steel_W_panel_zone":
            result_node = component.external_nodes.named_contents[
                "bottom_node"
            ]
            break
    return result_node


@dataclass(repr=False)
class TrussBarGenerator:
    """
    Introduces bar elements to a model.
    Bar elements are linear elements that can only carry axial load.

    """

    model: Model = field(repr=False)

    def add(
        self,
        xi_coord: float,
        yi_coord: float,
        lvl_key_i: int,
        offset_i: nparr,
        snap_i: str,
        xj_coord: float,
        yj_coord: float,
        lvl_key_j: int,
        offset_j: nparr,
        snap_j: str,
        transf_type: str,
        area: float,
        mat: UniaxialMaterial,
        outside_shape: Mesh,
        weight_per_length: float = 0.00,
        split_existing_i: bool = None,
        split_existing_j: bool = None,
        component_purpose: str = 'Truss Element',
    ) -> ComponentAssembly:
        """
        Adds a truss bar element.
        If offsets are required, they are implemented through the
        addition of RigidLink elements.

        """

        query = ElmQuery(self.model)
        ndg = NodeGenerator(self.model)
        lvls = self.model.levels

        lvl_i = lvls[lvl_key_i]

        sec_offset_global = np.zeros(3)

        node_i, eo_i = beam_placement_lookup(
            xi_coord,
            yi_coord,
            query,
            ndg,
            lvls,
            lvl_key_i,
            offset_i,
            sec_offset_global,
            split_existing_i,
            snap_i,
        )
        node_j, eo_j = beam_placement_lookup(
            xj_coord,
            yj_coord,
            query,
            ndg,
            lvls,
            lvl_key_j,
            offset_j,
            sec_offset_global,
            split_existing_j,
            snap_j,
        )

        # for braces, even if we specify a snap value that results in
        # the brace being connected to a different node (this is
        # typically done at panel zones), we still want the ends of
        # the brace to be located at the coordinates we specify, which
        # is different to what is done with the connectivity of beams.
        # Therefore, if the coordinates of the returned nodes differ
        # with those that we specified, we add the difference in the
        # offset, to move the ends of the brace back where we want
        # them to be. The effect of this is that the rigid offsets
        # (twonodelinks) will connect to that other node.
        i_diff = np.array((xi_coord, yi_coord)) - np.array(node_i.coords[0:2])
        if np.linalg.norm(i_diff) > common.EPSILON:
            eo_i[0:2] += i_diff
        j_diff = np.array((xj_coord, yj_coord)) - np.array(node_j.coords[0:2])
        if np.linalg.norm(j_diff) > common.EPSILON:
            eo_j[0:2] += j_diff

        # instantiate a component assembly
        component = ComponentAssembly(
            uid=self.model.uid_generator.new("component"),
            parent_collection=lvl_i.components,
            component_purpose=component_purpose,
        )
        # add it to the level
        lvl_i.components.add(component)
        # fill component assembly
        component.external_nodes.add(node_i)
        component.external_nodes.add(node_j)

        def prepare_connection(node_x: Node, eo_x: nparr) -> Node:
            """
            For each end of the bar element, creates a rigid link if
            an offset exists, and returns the node to which the bar
            element should connect to. This function is called twice,
            once for the i-end and once for the j-end.  For purposes
            of clarity, the index x will be used here, assuming that
            it will be substituted with i and j.
            """

            # if there is an offset at the x-end, create an internal node
            # and add a rigidlink element to the component assembly
            if np.linalg.norm(eo_x) > common.EPSILON:
                int_node_x = Node(
                    self.model.uid_generator.new('node'),
                    [*(np.array(node_x.coords) + eo_x)],
                )
                component.internal_nodes.add(int_node_x)
                n_x = int_node_x
                dirs, mats = zerolength_gen.fix_all(self.model)
                # flip the nodes if the element is about to be defined
                # upside down
                if (
                    np.allclose(
                        np.array(node_x.coords[0:2]),
                        np.array(int_node_x.coords[0:2]),
                    )
                    and int_node_x.coords[2] > node_x.coords[2]
                ):
                    x_axis, y_axis, _ = local_axes_from_points_and_angle(
                        np.array(int_node_x.coords),
                        np.array(node_x.coords),
                        0.00,
                    )
                    elm_link = TwoNodeLink(
                        component,
                        self.model.uid_generator.new("element"),
                        [node_x, int_node_x],
                        mats,
                        dirs,
                        x_axis,
                        y_axis,
                    )
                else:
                    x_axis, y_axis, _ = local_axes_from_points_and_angle(
                        np.array(node_x.coords),
                        np.array(int_node_x.coords),
                        0.00,
                    )
                    elm_link = TwoNodeLink(
                        component,
                        self.model.uid_generator.new("element"),
                        [node_x, int_node_x],
                        mats,
                        dirs,
                        x_axis,
                        y_axis,
                    )
                component.elements.add(elm_link)
            else:
                n_x = node_x
            return n_x

        # call the function here for the i and the j ends.
        n_i = prepare_connection(node_i, eo_i)
        n_j = prepare_connection(node_j, eo_j)

        # create the element
        elm_truss = TrussBar(
            parent_component=component,
            uid=self.model.uid_generator.new("element"),
            nodes=[n_i, n_j],
            transf_type=transf_type,
            area=area,
            mat=mat,
            outside_shape=outside_shape,
            weight_per_length=weight_per_length,
        )

        # add it to the component assembly
        component.elements.add(elm_truss)

        return component


@dataclass(repr=False)
class BeamColumnGenerator:
    """
    Introduces beamcolumn elements to a model.

    """

    model: Model = field(repr=False)

    def define_beamcolumn(
        self,
        assembly: ComponentAssembly,
        node_i: Node,
        node_j: Node,
        offset_i: nparr,
        offset_j: nparr,
        transf_type: str,
        section: ElasticSection | FiberSection,
        element_type: Type[Union[ElasticBeamColumn, DispBeamColumn]],
        angle: float = 0.00,
        n_x: Optional[float] = None,
        n_y: Optional[float] = None,
    ) -> ElasticBeamColumn | DispBeamColumn:
        """
        Adds a beamcolumn element to the model, connecting the
        specified nodes.

        """

        p_i = np.array(node_i.coords) + offset_i
        p_j = np.array(node_j.coords) + offset_j
        axes = local_axes_from_points_and_angle(
            p_i, p_j, angle
        )  # type: ignore
        if element_type.__name__ == "ElasticBeamColumn":
            assert isinstance(section, ElasticSection)
            transf = GeomTransf(
                transf_type,
                self.model.uid_generator.new("transformation"),
                offset_i,
                offset_j,
                *axes,
            )
            elm_el = ElasticBeamColumn(
                parent_component=assembly,
                uid=self.model.uid_generator.new("element"),
                nodes=[node_i, node_j],
                section=section,
                geomtransf=transf,
                n_x=n_x,
                n_y=n_y,
            )
            res: Union[ElasticBeamColumn, DispBeamColumn] = elm_el
        elif element_type.__name__ == "DispBeamColumn":
            assert isinstance(section, FiberSection)
            assert n_x is None
            assert n_y is None
            # TODO: add elastic section support
            transf = GeomTransf(
                transf_type,
                self.model.uid_generator.new("transformation"),
                offset_i,
                offset_j,
                *axes,
            )
            beam_integration = Lobatto(
                uid=self.model.uid_generator.new("beam integration"),
                parent_section=section,
                n_p=2,
            )
            elm_disp = DispBeamColumn(
                parent_component=assembly,
                uid=self.model.uid_generator.new("element"),
                nodes=[node_i, node_j],
                section=section,
                geomtransf=transf,
                integration=beam_integration,
            )
            res = elm_disp
        else:
            raise ValueError('Invalid element type: {element_type.__name__}')
        return res

    def define_zerolength(
        self,
        assembly: ComponentAssembly,
        node_i: Node,
        node_j: Node,
        x_axis: nparr,
        y_axis: nparr,
        zerolength_gen: Callable,  # type: ignore
        zerolength_gen_args: dict[str, object],
    ) -> ZeroLength:
        """
        Defines a zerolength element.

        """

        dirs, mats = zerolength_gen(model=self.model, **zerolength_gen_args)
        elm = ZeroLength(
            assembly,
            self.model.uid_generator.new("element"),
            [node_i, node_j],
            mats,
            dirs,
            x_axis,
            y_axis,
        )
        return elm

    def define_two_node_link(
        self,
        assembly: ComponentAssembly,
        node_i: Node,
        node_j: Node,
        x_axis: nparr,
        y_axis: nparr,
        zerolength_gen: Callable,  # type: ignore
        zerolength_gen_args: dict[str, object],
    ) -> TwoNodeLink:
        """
        Defines a TwoNodeLink element.

        """

        dirs, mats = zerolength_gen(model=self.model, **zerolength_gen_args)
        elm = TwoNodeLink(
            assembly,
            self.model.uid_generator.new("element"),
            [node_i, node_j],
            mats,
            dirs,
            x_axis,
            y_axis,
        )
        return elm

    def add_beamcolumn_elements_in_series(
        self,
        component,
        node_i,
        node_j,
        eo_i,
        eo_j,
        n_sub,
        transf_type,
        section,
        element_type,
        angle,
        camber_2,
        camber_3,
        n_x=None,
        n_y=None,
    ):
        """
        Adds beamcolumn elemens in series.

        """

        if (n_x is not None) or (n_y is not None):
            assert n_sub == 1

        if n_sub > 1:
            p_i = np.array(node_i.coords) + eo_i
            p_j = np.array(node_j.coords) + eo_j
            clear_len = np.linalg.norm(p_j - p_i)
            internal_pt_coords = np.linspace(
                tuple(p_i), tuple(p_j), num=n_sub + 1
            )

            # initial deformation
            t_vals = np.linspace(0.00, 1.00, num=n_sub + 1)
            # quadratic initial imperfection
            # offset_vals = 4.00 * (-t_vals**2 + t_vals)
            # sinusoidal initial imperfection
            offset_vals = np.sin(np.pi * t_vals)
            offset_2 = offset_vals * camber_2 * clear_len
            offset_3 = offset_vals * camber_3 * clear_len
            camber_offset: nparr = np.column_stack(
                (np.zeros(n_sub + 1), offset_2, offset_3)
            )
            x_axis, y_axis, z_axis = local_axes_from_points_and_angle(
                p_i, p_j, angle
            )
            t_glob_to_loc = transformation_matrix(x_axis, y_axis, z_axis)
            t_loc_to_glob = t_glob_to_loc.T
            camber_offset_global = (t_loc_to_glob @ camber_offset.T).T
            internal_pt_coords += camber_offset_global

            intnodes = []
            for i in range(1, len(internal_pt_coords) - 1):
                intnode = Node(
                    self.model.uid_generator.new("node"),
                    [*internal_pt_coords[i]],
                )
                component.internal_nodes.add(intnode)
                intnodes.append(intnode)
        for i in range(n_sub):
            if i == 0:
                n_i = node_i
                o_i = eo_i
            else:
                n_i = intnodes[i - 1]
                o_i = np.zeros(3)
            if i == n_sub - 1:
                n_j = node_j
                o_j = eo_j
            else:
                n_j = intnodes[i]
                o_j = np.zeros(3)
            if element_type.__name__ not in {
                'ElasticBeamColumn',
                'DispBeamColumn',
            }:
                raise TypeError(
                    f"Unsupported element type: {element_type.__name__}"
                )
            element = self.define_beamcolumn(
                assembly=component,
                node_i=n_i,
                node_j=n_j,
                offset_i=o_i,
                offset_j=o_j,
                transf_type=transf_type,
                section=section,
                element_type=element_type,
                angle=angle,
                n_x=n_x,
                n_y=n_y,
            )
            component.elements.add(element)

    def generate_plain_component_assembly(
        self,
        component_purpose,
        lvl,
        node_i,
        node_j,
        n_sub,
        eo_i,
        eo_j,
        section,
        element_type,
        transf_type,
        angle,
        camber_2,
        camber_3,
        n_x=None,
        n_y=None,
    ):
        """
        Generates a plain component assembly, with line elements in
        series.

        """

        assert isinstance(node_i, Node)
        assert isinstance(node_j, Node)

        uids = [node.uid for node in (node_i, node_j)]
        uids.sort()
        uids_tuple = (*uids,)
        assert uids_tuple not in self.model.component_connectivity()

        # instantiate a component assembly
        component = ComponentAssembly(
            uid=self.model.uid_generator.new("component"),
            parent_collection=lvl.components,
            component_purpose=component_purpose,
        )
        # add it to the level
        lvl.components.add(component)
        # fill component assembly
        component.external_nodes.add(node_i)
        component.external_nodes.add(node_j)

        self.add_beamcolumn_elements_in_series(
            component,
            node_i,
            node_j,
            eo_i,
            eo_j,
            n_sub,
            transf_type,
            section,
            element_type,
            angle,
            camber_2,
            camber_3,
            n_x,
            n_y,
        )

        return component

    def generate_hinged_component_assembly(
        self,
        component_purpose,
        lvl,
        node_i,
        node_j,
        n_sub,
        eo_i,
        eo_j,
        section,
        element_type,
        transf_type,
        angle,
        camber_2,
        camber_3,
        n_x,
        n_y,
        zerolength_gen_i,
        zerolength_gen_args_i,
        zerolength_gen_j,
        zerolength_gen_args_j,
    ):
        """
        Defines a component assembly that is comprised of
        beamcolumn elements connected in series with nonlinear springs
        attached at the ends, followed by another sequence of
        beamcolumn elements (in order to be able to specify rigid offsets).

        """

        uids = [node.uid for node in (node_i, node_j)]
        uids.sort()
        uids_tuple = (*uids,)
        assert uids_tuple not in self.model.component_connectivity()

        # instantiate a component assembly
        component = ComponentAssembly(
            uid=self.model.uid_generator.new("component"),
            parent_collection=lvl.components,
            component_purpose=component_purpose,
        )
        # fill component assembly
        component.external_nodes.add(node_i)
        component.external_nodes.add(node_j)
        # add it to the level
        lvl.components.add(component)

        p_i = np.array(node_i.coords) + eo_i
        p_j = np.array(node_j.coords) + eo_j
        axes = local_axes_from_points_and_angle(p_i, p_j, angle)
        x_axis, y_axis, _ = axes
        clear_length = np.linalg.norm(p_j - p_i)
        zerolength_gen_args_i.update({"element_length": clear_length})
        zerolength_gen_args_j.update({"element_length": clear_length})

        # we can have hinges at both ends, or just one of the two ends.
        # ...or even no hinges!
        if zerolength_gen_i:
            hinge_location_i = p_i + x_axis * zerolength_gen_args_i["distance"]
            nh_i_out = Node(
                self.model.uid_generator.new("node"), [*hinge_location_i]
            )
            nh_i_in = Node(
                self.model.uid_generator.new("node"), [*hinge_location_i]
            )
            nh_i_in.visibility.connected_to_zerolength = True
            component.internal_nodes.add(nh_i_out)
            component.internal_nodes.add(nh_i_in)
            if "element_type" in zerolength_gen_args_i:
                element_type_i = zerolength_gen_args_i["element_type"]
            else:
                element_type_i = element_type
            if element_type_i.__name__ in (
                "ElasticBeamColumn",
                "DispBeamColumn",
            ):
                if "section" in zerolength_gen_args_i:
                    section_i = zerolength_gen_args_i["section"]
                else:
                    section_i = section
                if "transf_type" in zerolength_gen_args_i:
                    transf_type_i = zerolength_gen_args_i["transf_type"]
                else:
                    transf_type_i = transf_type
                self.add_beamcolumn_elements_in_series(
                    component,
                    node_i,
                    nh_i_out,
                    eo_i,
                    np.zeros(3),
                    zerolength_gen_args_i["n_sub"],
                    transf_type_i,
                    section_i,
                    element_type_i,
                    angle,
                    0.00,
                    0.00,
                )
            elif element_type_i.__name__ == "TwoNodeLink":
                elm = self.define_two_node_link(
                    component,
                    node_i,
                    nh_i_out,
                    x_axis,
                    y_axis,
                    zerolength_gen.fix_all,
                    {},
                )
                component.elements.add(elm)
            else:
                raise ValueError(f"Invalid element_type_i: {element_type_i}")
            zerolen_elm = self.define_zerolength(
                component,
                nh_i_out,
                nh_i_in,
                x_axis,
                y_axis,
                zerolength_gen_i,
                zerolength_gen_args_i,
            )
            component.elements.add(zerolen_elm)
            conn_node_i = nh_i_in
            conn_eo_i = np.zeros(3)
        else:
            conn_node_i = node_i
            conn_eo_i = eo_i
        if zerolength_gen_j:
            hinge_location_j = p_i + x_axis * (
                clear_length - zerolength_gen_args_j["distance"]
            )
            nh_j_out = Node(
                self.model.uid_generator.new("node"), [*hinge_location_j]
            )
            nh_j_in = Node(
                self.model.uid_generator.new("node"), [*hinge_location_j]
            )
            nh_j_in.visibility.connected_to_zerolength = True
            component.internal_nodes.add(nh_j_out)
            component.internal_nodes.add(nh_j_in)
            if "element_type" in zerolength_gen_args_j:
                element_type_j = zerolength_gen_args_j["element_type"]
            else:
                element_type_j = element_type
            if element_type_j.__name__ in (
                "ElasticBeamColumn",
                "DispBeamColumn",
            ):
                if "section" in zerolength_gen_args_j:
                    section_j = zerolength_gen_args_j["section"]
                else:
                    section_j = section
                if "transf_type" in zerolength_gen_args_j:
                    transf_type_j = zerolength_gen_args_j["transf_type"]
                else:
                    transf_type_j = transf_type
                self.add_beamcolumn_elements_in_series(
                    component,
                    nh_j_out,
                    node_j,
                    np.zeros(3),
                    eo_j,
                    zerolength_gen_args_j["n_sub"],
                    transf_type_j,
                    section_j,
                    element_type_j,
                    angle,
                    0.00,
                    0.00,
                )
            elif element_type_j.__name__ == "TwoNodeLink":
                elm = self.define_two_node_link(
                    component,
                    nh_j_out,
                    node_j,
                    x_axis,
                    y_axis,
                    zerolength_gen.fix_all,
                    {},
                )
                component.elements.add(elm)
            else:
                raise ValueError(f"Invalid element_type_j: {element_type_j}")
            zerolen_elm = self.define_zerolength(
                component,
                nh_j_out,
                nh_j_in,
                -x_axis,
                y_axis,
                zerolength_gen_j,
                zerolength_gen_args_j,
            )
            component.elements.add(zerolen_elm)
            conn_node_j = nh_j_in
            conn_eo_j = np.zeros(3)
        else:
            conn_node_j = node_j
            conn_eo_j = eo_j

        self.add_beamcolumn_elements_in_series(
            component,
            conn_node_i,
            conn_node_j,
            conn_eo_i,
            conn_eo_j,
            n_sub,
            transf_type,
            section,
            element_type,
            angle,
            camber_2,
            camber_3,
            n_x,
            n_y,
        )
        return component

    def add_vertical_active(
        self,
        x_coord: float,
        y_coord: float,
        offset_i: nparr,
        offset_j: nparr,
        transf_type: str,
        n_sub: int,
        section: ElasticSection | FiberSection,
        element_type: Type[Union[ElasticBeamColumn, DispBeamColumn]],
        placement: str = "centroid",
        angle: float = 0.00,
        camber_2: float = 0.00,
        camber_3: float = 0.00,
        method: str = "generate_plain_component_assembly",
        additional_args: dict[str, object] = {},
    ) -> dict[int, ComponentAssembly]:
        """
        Adds a vertical beamcolumn element to all active levels.  This
        method assumes that the levels are defined in order, from
        lowest to highest elevation, with consecutive ascending
        integer keys.

        """

        ndg = NodeGenerator(self.model)
        query = ElmQuery(self.model)
        lvls = self.model.levels
        assert lvls.active, "No active levels."
        defined_component_assemblies: dict[int, ComponentAssembly] = {}
        for key in lvls.active:
            lvl = lvls[key]
            if key - 1 not in lvls:
                continue

            top_node = query.search_node_lvl(x_coord, y_coord, key)
            if not top_node:
                top_node = ndg.add_node_lvl(x_coord, y_coord, key)

            bottom_node = query.search_node_lvl(x_coord, y_coord, key - 1)
            if not bottom_node:
                bottom_node = ndg.add_node_lvl(x_coord, y_coord, key - 1)

            # check for a panel zone
            top_node = look_for_panel_zone(top_node, lvl, query)

            p_i = np.array(top_node.coords) + offset_i
            p_j = np.array(bottom_node.coords) + offset_j
            sec_offset_global = retrieve_snap_pt_global_offset(
                placement, section, p_i, p_j, angle
            )
            p_i += sec_offset_global
            p_j += sec_offset_global
            eo_i = offset_i + sec_offset_global
            eo_j = offset_j + sec_offset_global

            args = {
                "component_purpose": "vertical_component",
                "lvl": lvl,
                "node_i": top_node,
                "node_j": bottom_node,
                "n_sub": n_sub,
                "eo_i": eo_i,
                "eo_j": eo_j,
                "section": section,
                "element_type": element_type,
                "transf_type": transf_type,
                "angle": angle,
                "camber_2": camber_2,
                "camber_3": camber_3,
            }

            args.update(additional_args)
            assert hasattr(self, method), f"Method not available: {method}"
            mthd = getattr(self, method)
            defined_component_assemblies[key] = mthd(**args)
        return defined_component_assemblies

    def add_horizontal_active(
        self,
        xi_coord: float,
        yi_coord: float,
        xj_coord: float,
        yj_coord: float,
        offset_i: nparr,
        offset_j: nparr,
        snap_i: str,
        snap_j: str,
        transf_type: str,
        n_sub: int,
        section: ElasticSection,
        element_type: Type[Union[ElasticBeamColumn, DispBeamColumn]],
        placement: str = "centroid",
        angle: float = 0.00,
        camber_2: float = 0.00,
        camber_3: float = 0.00,
        split_existing_i: Optional[ComponentAssembly] = None,
        split_existing_j: Optional[ComponentAssembly] = None,
        h_offset_i: float = 0.00,
        h_offset_j: float = 0.00,
        method: str = "generate_plain_component_assembly",
        additional_args: dict[str, object] = {},
    ) -> dict[int, ComponentAssembly]:
        """
        Adds a diagonal beamcolumn element to all active levels.

        """

        query = ElmQuery(self.model)
        ndg = NodeGenerator(self.model)
        lvls = self.model.levels
        assert lvls.active, "No active levels."
        defined_component_assemblies: dict[int, ComponentAssembly] = {}
        for key in lvls.active:
            lvl = lvls[key]
            lvl_prev = lvls.get(key - 1)

            if not lvl_prev:
                continue

            p_i_init = np.array((xi_coord, yi_coord, lvl.elevation)) + offset_i
            p_j_init = np.array((xj_coord, yj_coord, lvl.elevation)) + offset_j

            # retrieve local coordinate system
            x_axis, y_axis, z_axis = local_axes_from_points_and_angle(
                p_i_init, p_j_init, angle
            )  # type: ignore

            p_i_init += h_offset_i * x_axis
            p_j_init += -h_offset_j * x_axis
            offset_i += h_offset_i * x_axis
            offset_j += -h_offset_j * x_axis

            if section.snap_points and (placement != "centroid"):
                # obtain offset from section (local system)
                d_z, d_y = section.snap_points[placement]
                sec_offset_local: nparr = np.array([0.00, d_y, d_z])
                t_glob_to_loc = transformation_matrix(x_axis, y_axis, z_axis)
                t_loc_to_glob = t_glob_to_loc.T
                sec_offset_global = t_loc_to_glob @ sec_offset_local
            else:
                sec_offset_global = np.zeros(3)

            node_i, eo_i = beam_placement_lookup(
                xi_coord,
                yi_coord,
                query,
                ndg,
                lvls,
                key,
                offset_i,
                sec_offset_global,
                split_existing_i,
                snap_i,
            )
            node_j, eo_j = beam_placement_lookup(
                xj_coord,
                yj_coord,
                query,
                ndg,
                lvls,
                key,
                offset_j,
                sec_offset_global,
                split_existing_j,
                snap_j,
            )

            args = {
                "component_purpose": "horizontal_component",
                "lvl": lvl,
                "node_i": node_i,
                "node_j": node_j,
                "n_sub": n_sub,
                "eo_i": eo_i,
                "eo_j": eo_j,
                "section": section,
                "element_type": element_type,
                "transf_type": transf_type,
                "angle": angle,
                "camber_2": camber_2,
                "camber_3": camber_3,
            }

            args.update(additional_args)
            assert hasattr(self, method), f"Method not available: {method}"
            mthd = getattr(self, method)
            defined_component_assemblies[key] = mthd(**args)
        return defined_component_assemblies

    def add_diagonal_active(
        self,
        xi_coord: float,
        yi_coord: float,
        xj_coord: float,
        yj_coord: float,
        offset_i: nparr,
        offset_j: nparr,
        snap_i: str,
        snap_j: str,
        transf_type: str,
        n_sub: int,
        section: ElasticSection,
        element_type: Type[Union[ElasticBeamColumn, DispBeamColumn]],
        placement: str = 'centroid',
        angle: float = 0.00,
        camber_2: float = 0.00,
        camber_3: float = 0.00,
        split_existing_i: Optional[ComponentAssembly] = None,
        split_existing_j: Optional[ComponentAssembly] = None,
        method: str = 'generate_plain_component_assembly',
        additional_args: dict[str, object] = {},
    ) -> dict[int, ComponentAssembly]:
        """
        Adds a diagonal beamcolumn element to all active levels.

        """

        query = ElmQuery(self.model)
        ndg = NodeGenerator(self.model)
        lvls = self.model.levels
        assert lvls.active, 'No active levels.'
        defined_component_assemblies: dict[int, ComponentAssembly] = {}
        for key in lvls.active:
            lvl = lvls[key]
            lvl_prev = lvls.get(key - 1)

            if not lvl_prev:
                continue

            p_i_init = np.array((xi_coord, yi_coord, lvl.elevation)) + offset_i
            p_j_init = np.array((xj_coord, yj_coord, lvl.elevation)) + offset_j

            if section.snap_points and (placement != 'centroid'):
                # obtain offset from section (local system)
                d_z, d_y = section.snap_points[placement]
                sec_offset_local: nparr = np.array([0.00, d_y, d_z])
                # retrieve local coordinate system
                x_axis, y_axis, z_axis = local_axes_from_points_and_angle(
                    p_i_init, p_j_init, angle
                )  # type: ignore
                t_glob_to_loc = transformation_matrix(x_axis, y_axis, z_axis)
                t_loc_to_glob = t_glob_to_loc.T
                sec_offset_global = t_loc_to_glob @ sec_offset_local
            else:
                sec_offset_global = np.zeros(3)

            node_i, eo_i = beam_placement_lookup(
                xi_coord,
                yi_coord,
                query,
                ndg,
                lvls,
                key,
                offset_i,
                sec_offset_global,
                split_existing_i,
                snap_i,
            )
            node_j, eo_j = beam_placement_lookup(
                xj_coord,
                yj_coord,
                query,
                ndg,
                lvls,
                key - 1,
                offset_j,
                sec_offset_global,
                split_existing_j,
                snap_j,
            )

            args = {
                'component_purpose': 'diagonal_component',
                'lvl': lvl,
                'node_i': node_i,
                'node_j': node_j,
                'n_sub': n_sub,
                'eo_i': eo_i,
                'eo_j': eo_j,
                'section': section,
                'element_type': element_type,
                'transf_type': transf_type,
                'angle': angle,
                'camber_2': camber_2,
                'camber_3': camber_3,
            }

            args.update(additional_args)
            assert hasattr(self, method), f'Method not available: {method}'
            mthd = getattr(self, method)
            defined_component_assemblies[key] = mthd(**args)
        return defined_component_assemblies

    def add_pz_active(
        self,
        x_coord: float,
        y_coord: float,
        section: ElasticSection,
        physical_material: PhysicalMaterial,
        angle: float,
        column_depth: float,
        beam_depth: float,
        zerolength_method: str,
        zerolength_args: dict[str, object],
    ) -> dict[int, ComponentAssembly]:
        """
        Adds a component assembly representing a steel W-section
        panel zone joint.

        """

        ndg = NodeGenerator(self.model)
        query = ElmQuery(self.model)
        lvls = self.model.levels
        assert lvls.active, "No active levels."
        defined_components: dict[int, ComponentAssembly] = {}
        for key in lvls.active:
            lvl = lvls[key]
            if key - 1 not in lvls:
                continue

            top_node = query.search_node_lvl(x_coord, y_coord, key)
            if not top_node:
                top_node = ndg.add_node_lvl(x_coord, y_coord, key)

            # instantiate a component assembly
            component = ComponentAssembly(
                uid=self.model.uid_generator.new("component"),
                parent_collection=lvl.components,
                component_purpose="steel_W_panel_zone",
            )
            # add it to the level
            lvl.components.add(component)

            p_i: nparr = np.array(top_node.coords)
            p_j = np.array(top_node.coords) + np.array(
                (0.00, 0.00, -beam_depth)
            )
            x_axis, y_axis, z_axis = local_axes_from_points_and_angle(
                p_i, p_j, angle
            )  # type: ignore

            # determine node locations
            top_h_f_loc = p_i + y_axis * column_depth / 2.00
            top_h_b_loc = p_i - y_axis * column_depth / 2.00
            top_v_f_loc = p_i + y_axis * column_depth / 2.00
            top_v_b_loc = p_i - y_axis * column_depth / 2.00
            mid_v_f_loc = (
                p_i + y_axis * column_depth / 2.00 + x_axis * beam_depth / 2.00
            )
            mid_v_b_loc = (
                p_i - y_axis * column_depth / 2.00 + x_axis * beam_depth / 2.00
            )
            bottom_h_f_loc = (
                p_i + y_axis * column_depth / 2.00 + x_axis * beam_depth
            )
            bottom_h_b_loc = (
                p_i - y_axis * column_depth / 2.00 + x_axis * beam_depth
            )
            bottom_v_f_loc = (
                p_i + y_axis * column_depth / 2.00 + x_axis * beam_depth
            )
            bottom_v_b_loc = (
                p_i - y_axis * column_depth / 2.00 + x_axis * beam_depth
            )

            # define nodes
            top_h_f = Node(
                self.model.uid_generator.new("node"), [*top_h_f_loc]
            )
            top_h_b = Node(
                self.model.uid_generator.new("node"), [*top_h_b_loc]
            )
            top_v_f = Node(
                self.model.uid_generator.new("node"), [*top_v_f_loc]
            )
            top_v_f.visibility.connected_to_zerolength = True
            top_v_b = Node(
                self.model.uid_generator.new("node"), [*top_v_b_loc]
            )
            top_v_b.visibility.connected_to_zerolength = True

            mid_v_f = ndg.add_node_lvl_xyz(
                mid_v_f_loc[0], mid_v_f_loc[1], mid_v_f_loc[2], lvl.uid
            )
            mid_v_b = ndg.add_node_lvl_xyz(
                mid_v_b_loc[0], mid_v_b_loc[1], mid_v_b_loc[2], lvl.uid
            )

            bottom_h_f = Node(
                self.model.uid_generator.new("node"), [*bottom_h_f_loc]
            )
            bottom_h_b = Node(
                self.model.uid_generator.new("node"), [*bottom_h_b_loc]
            )
            bottom_v_f = Node(
                self.model.uid_generator.new("node"), [*bottom_v_f_loc]
            )
            bottom_v_f.visibility.connected_to_zerolength = True
            bottom_v_b = Node(
                self.model.uid_generator.new("node"), [*bottom_v_b_loc]
            )
            bottom_v_b.visibility.connected_to_zerolength = True

            bottom_mid = ndg.add_node_lvl_xyz(p_j[0], p_j[1], p_j[2], lvl.uid)

            factor_i = 1.00e1
            factor_a = 1.00e1
            factor_j = 1.00e1

            new_uid = self.model.uid_generator.new("section")
            rigid_sec = ElasticSection(
                name="rigid_link_section",
                uid=new_uid,
                outside_shape=None,
                snap_points=None,
                e_mod=section.e_mod,
                area=section.area * factor_a,
                i_y=section.i_y * factor_i,
                i_x=section.i_x * factor_i,
                g_mod=section.g_mod,
                j_mod=section.j_mod * factor_j,
                sec_w=0.00,
            )

            elm_top_h_f = ElasticBeamColumn(
                component,
                self.model.uid_generator.new("element"),
                [top_node, top_h_f],
                rigid_sec,
                GeomTransf(
                    "Linear",
                    self.model.uid_generator.new("transformation"),
                    np.zeros(3),
                    np.zeros(3),
                    y_axis,
                    -x_axis,
                    z_axis,
                ),
            )
            elm_top_h_f.visibility.hidden_when_extruded = True
            elm_top_h_f.visibility.hidden_basic_forces = True

            elm_top_h_b = ElasticBeamColumn(
                component,
                self.model.uid_generator.new("element"),
                [top_h_b, top_node],
                rigid_sec,
                GeomTransf(
                    "Linear",
                    self.model.uid_generator.new("transformation"),
                    np.zeros(3),
                    np.zeros(3),
                    y_axis,
                    -x_axis,
                    z_axis,
                ),
            )
            elm_top_h_b.visibility.hidden_when_extruded = True
            elm_top_h_b.visibility.hidden_basic_forces = True

            elm_bottom_h_f = ElasticBeamColumn(
                component,
                self.model.uid_generator.new("element"),
                [bottom_mid, bottom_h_f],
                rigid_sec,
                GeomTransf(
                    "Linear",
                    self.model.uid_generator.new("transformation"),
                    np.zeros(3),
                    np.zeros(3),
                    y_axis,
                    -x_axis,
                    z_axis,
                ),
            )
            elm_bottom_h_f.visibility.hidden_when_extruded = True
            elm_bottom_h_f.visibility.hidden_basic_forces = True

            elm_bottom_h_b = ElasticBeamColumn(
                component,
                self.model.uid_generator.new("element"),
                [bottom_h_b, bottom_mid],
                rigid_sec,
                GeomTransf(
                    "Linear",
                    self.model.uid_generator.new("transformation"),
                    np.zeros(3),
                    np.zeros(3),
                    y_axis,
                    -x_axis,
                    z_axis,
                ),
            )
            elm_bottom_h_b.visibility.hidden_when_extruded = True
            elm_bottom_h_b.visibility.hidden_basic_forces = True

            elm_top_v_f = ElasticBeamColumn(
                component,
                self.model.uid_generator.new("element"),
                [top_v_f, mid_v_f],
                rigid_sec,
                GeomTransf(
                    "Linear",
                    self.model.uid_generator.new("transformation"),
                    np.zeros(3),
                    np.zeros(3),
                    x_axis,
                    y_axis,
                    z_axis,
                ),
            )
            elm_top_v_f.visibility.hidden_when_extruded = True
            elm_top_v_f.visibility.hidden_basic_forces = True

            elm_top_v_b = ElasticBeamColumn(
                component,
                self.model.uid_generator.new("element"),
                [top_v_b, mid_v_b],
                rigid_sec,
                GeomTransf(
                    "Linear",
                    self.model.uid_generator.new("transformation"),
                    np.zeros(3),
                    np.zeros(3),
                    x_axis,
                    y_axis,
                    z_axis,
                ),
            )
            elm_top_v_b.visibility.hidden_when_extruded = True
            elm_top_v_b.visibility.hidden_basic_forces = True

            elm_bottom_v_f = ElasticBeamColumn(
                component,
                self.model.uid_generator.new("element"),
                [mid_v_f, bottom_v_f],
                rigid_sec,
                GeomTransf(
                    "Linear",
                    self.model.uid_generator.new("transformation"),
                    np.zeros(3),
                    np.zeros(3),
                    x_axis,
                    y_axis,
                    z_axis,
                ),
            )
            elm_bottom_v_f.visibility.hidden_when_extruded = True
            elm_bottom_v_f.visibility.hidden_basic_forces = True

            elm_bottom_v_b = ElasticBeamColumn(
                component,
                self.model.uid_generator.new("element"),
                [mid_v_b, bottom_v_b],
                rigid_sec,
                GeomTransf(
                    "Linear",
                    self.model.uid_generator.new("transformation"),
                    np.zeros(3),
                    np.zeros(3),
                    x_axis,
                    y_axis,
                    z_axis,
                ),
            )
            elm_bottom_v_b.visibility.hidden_when_extruded = True
            elm_bottom_v_b.visibility.hidden_basic_forces = True

            elm_interior = ElasticBeamColumn(
                component,
                self.model.uid_generator.new("element"),
                [top_node, bottom_mid],
                section,
                GeomTransf(
                    "Linear",
                    self.model.uid_generator.new("transformation"),
                    np.zeros(3),
                    np.zeros(3),
                    x_axis,
                    y_axis,
                    z_axis,
                ),
            )
            elm_interior.visibility.skip_opensees_definition = True
            elm_interior.visibility.hidden_at_line_plots = True

            assert hasattr(
                zerolength_gen, zerolength_method
            ), f"Method not available: {zerolength_method}"
            mthd = getattr(zerolength_gen, zerolength_method)

            # define zerolength elements
            zerolength_gen_args = {
                "section": section,
                "physical_material": physical_material,
                "pz_length": beam_depth,
            }
            zerolength_gen_args.update(zerolength_args)
            zerolen_top_f = self.define_zerolength(
                component,
                top_h_f,
                top_v_f,
                x_axis,
                y_axis,
                mthd,
                zerolength_gen_args,
            )
            zerolen_top_b = self.define_zerolength(
                component,
                top_h_b,
                top_v_b,
                x_axis,
                y_axis,
                zerolength_gen.release_6,
                {},
            )
            zerolen_bottom_f = self.define_zerolength(
                component,
                bottom_h_f,
                bottom_v_f,
                x_axis,
                y_axis,
                zerolength_gen.release_6,
                {},
            )
            zerolen_bottom_b = self.define_zerolength(
                component,
                bottom_h_b,
                bottom_v_b,
                x_axis,
                y_axis,
                zerolength_gen.release_6,
                {},
            )

            # fill component assembly
            component.external_nodes.add(top_node)
            component.external_nodes.named_contents["top_node"] = top_node
            component.external_nodes.add(bottom_mid)
            component.external_nodes.named_contents["bottom_node"] = bottom_mid
            component.external_nodes.add(mid_v_f)
            component.external_nodes.named_contents["middle_front"] = mid_v_f
            component.external_nodes.add(mid_v_b)
            component.external_nodes.named_contents["middle_back"] = mid_v_b

            component.internal_nodes.add(top_h_f)
            component.internal_nodes.add(top_h_b)
            component.internal_nodes.add(top_v_f)
            component.internal_nodes.add(top_v_b)
            component.internal_nodes.add(bottom_h_f)
            component.internal_nodes.add(bottom_h_b)
            component.internal_nodes.add(bottom_v_f)
            component.internal_nodes.add(bottom_v_b)

            component.elements.add(elm_top_h_f)
            (component.elements.named_contents["elm_top_h_f"]) = elm_top_h_f
            component.elements.add(elm_top_h_b)
            (component.elements.named_contents["elm_top_h_b"]) = elm_top_h_b
            component.elements.add(elm_bottom_h_f)
            component.elements.add(elm_bottom_h_b)
            component.elements.add(elm_top_v_f)
            component.elements.add(elm_top_v_b)
            component.elements.add(elm_bottom_v_f)
            component.elements.add(elm_bottom_v_b)
            component.elements.add(elm_interior)
            (component.elements.named_contents["elm_interior"]) = elm_interior

            component.elements.add(zerolen_top_f)
            (
                component.elements.named_contents["nonlinear_spring"]
            ) = zerolen_top_f  # type: ignore
            component.elements.add(zerolen_top_b)
            component.elements.add(zerolen_bottom_f)
            component.elements.add(zerolen_bottom_b)
            defined_components[key] = component

        return defined_components



"""
Defines the :func:`~split_component` method.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import numpy as np
import numpy.typing as npt
from ..line import Line
from ..ops.element import ElasticBeamColumn
from ..ops.element import DispBeamColumn
from ..ops.element import GeomTransf
from ..ops.section import ElasticSection
from ..ops.section import FiberSection
from ..ops.node import Node
from ..component_assembly import ComponentAssembly
from .. import common
from ..ops.element import Lobatto


nparr = npt.NDArray[np.float64]


def split_component(
    component: ComponentAssembly, point: nparr
) -> tuple[Node, nparr]:
    """
    Splits a beam-functioning component assembly to accomodate for a
    node required to connect another component assembly.

    Arguments:
      component: Component assembly to split.
      point: Point to be used to split the component.

    WARNING:
      This function might no longer work. Needs to be tested and
      perhaps updated.

    """

    uid_generator = (
        component.parent_collection.parent.parent_model.uid_generator
    )

    elms = []
    elms.extend(
        [
            elm
            for elm in component.elements.values()
            if isinstance(elm, (ElasticBeamColumn, DispBeamColumn))
        ]
    )
    distances = np.zeros(len(elms))
    for i, elm in enumerate(elms):
        p_i = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
        p_j = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
        line = Line("", p_i, p_j)
        dist = line.point_distance(point)
        distances[i] = dist
    np.nan_to_num(distances, copy=False, nan=np.inf)
    i_min = np.argmin(distances)
    closest_elm = elms[i_min]
    p_i = (
        np.array(closest_elm.nodes[0].coords) + closest_elm.geomtransf.offset_i
    )
    p_j = (
        np.array(closest_elm.nodes[1].coords) + closest_elm.geomtransf.offset_j
    )
    line = Line("", p_i, p_j)
    split_point = line.project(point)
    assert split_point is not None  # check if it exists

    # first check if a node already exists there
    inodes = component.internal_nodes.values()
    for inode in inodes:
        if (
            np.linalg.norm(np.array(inode.coords) - split_point)
            < common.EPSILON
        ):
            avail_node = inode
            offset = point - np.array(avail_node.coords)
            return avail_node, offset

    # otherwise:

    # remove existing line element
    node_i = closest_elm.nodes[0]
    node_j = closest_elm.nodes[1]
    prev_section = closest_elm.section
    prev_gtransf = closest_elm.geomtransf
    component.elements.pop(closest_elm.uid)

    # add split node
    middle_node = Node(
        uid_generator.new(
            "node"
        ),
        list(split_point),
    )
    component.internal_nodes.add(middle_node)
    # add two new line elements
    # part i
    o_i = prev_gtransf.offset_i
    o_j = np.zeros(3)
    n_i = node_i
    n_j = middle_node
    transf_i = GeomTransf(
        prev_gtransf.transf_type,
        uid_generator.new(
            "transformation"
        ),
        o_i,
        o_j,
        prev_gtransf.x_axis,
        prev_gtransf.y_axis,
        prev_gtransf.z_axis,
    )
    if isinstance(closest_elm, ElasticBeamColumn):
        assert isinstance(prev_section, ElasticSection)
        elm_i = ElasticBeamColumn(
            component,
            uid_generator.new(
                "element"
            ),
            [n_i, n_j],
            prev_section,
            transf_i,
        )
        component.elements.add(elm_i)
    elif isinstance(closest_elm, DispBeamColumn):
        assert isinstance(closest_elm.integration, Lobatto)
        beam_integration = Lobatto(
            uid=(
                uid_generator.new(
                    "beam integration"
                )
            ),
            parent_section=prev_section,
            n_p=closest_elm.integration.n_p,
        )
        assert isinstance(prev_section, FiberSection)
        elm_i = DispBeamColumn(  # type: ignore
            component,
            (
                uid_generator.new(
                    "element"
                )
            ),
            [n_i, n_j],
            prev_section,
            transf_i,
            beam_integration,
        )
        component.elements.add(elm_i)
    # part j
    o_i = np.zeros(3)
    o_j = prev_gtransf.offset_j
    n_i = middle_node
    n_j = node_j
    transf_j = GeomTransf(
        prev_gtransf.transf_type,
        (
            uid_generator.new(
                "transformation"
            )
        ),
        o_i,
        o_j,
        prev_gtransf.x_axis,
        prev_gtransf.y_axis,
        prev_gtransf.z_axis,
    )
    if isinstance(closest_elm, ElasticBeamColumn):
        assert isinstance(prev_section, ElasticSection)
        elm_j = ElasticBeamColumn(
            component,
            (
                uid_generator.new(
                    "element"
                )
            ),
            [n_i, n_j],
            prev_section,
            transf_j,
        )
        component.elements.add(elm_j)
    elif isinstance(closest_elm, DispBeamColumn):
        assert isinstance(closest_elm.integration, Lobatto)
        beam_integration = Lobatto(
            uid=(
                uid_generator.new(
                    "beam integration"
                )
            ),
            parent_section=prev_section,
            n_p=closest_elm.integration.n_p,
        )
        assert isinstance(prev_section, FiberSection)
        elm_j = DispBeamColumn(  # type: ignore
            component,
            (
                uid_generator.new(
                    "element"
                )
            ),
            [n_i, n_j],
            prev_section,
            transf_j,
            beam_integration,
        )
        component.elements.add(elm_j)

    # calculate offset and return
    offset = point - np.array(middle_node.coords)
    return middle_node, offset



"""
Model Generator for OpenSees ~ level
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from dataclasses import dataclass, field
from .obj_collections import Collection
from .obj_collections import NodeCollection

if TYPE_CHECKING:
    from .model import Model
    from .component_assembly import ComponentAssembly


@dataclass
class Level:
    """
    Level Object. Levels are part of a model and they contain primary
      nodes and component assemblies.

    Attributes:
        parent_model (Model)
        uid (int)
        elevation (float)
        nodes (NodeCollection)
        components (Collection)

    Example:
        >>> from osmg.model import Model
        >>> model = Model(name='example_model')
        >>> level = Level(parent_model=model, uid=1, elevation=0.0)
        >>> level.parent_model.name
        'example_model'
        >>> level.uid
        1
        >>> level.elevation
        0.0
        >>> type(level.nodes)
        <class 'osmg.obj_collections.NodeCollection'>
        >>> type(level.components)
        <class 'osmg.obj_collections.Collection'>
    """

    parent_model: Model = field(repr=False)
    uid: int
    elevation: float
    nodes: NodeCollection = field(init=False, repr=False)
    components: Collection[int, ComponentAssembly] = field(
        init=False, repr=False
    )

    def __post_init__(self):
        self.nodes = NodeCollection(self)
        self.components = Collection(self)

    def __repr__(self):
        res = ""
        res += "Level object\n"
        res += f"parent_model: {self.parent_model.name}\n"
        res += f"uid: {self.uid}\n"
        res += f"elevation: {self.elevation}\n"
        res += "Nodes: \n"
        res += self.nodes.__srepr__() + "\n"
        res += "Components: \n"
        res += self.components.__srepr__() + "\n"
        return res



"""
Defines PhysicalMaterial objects.
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from dataclasses import dataclass


@dataclass
class PhysicalMaterial:
    """
    Physical material.
    We use this for self-weight, plotting enhancements etc.

    Example:
        >>> mat1 = PhysicalMaterial(
        ...     1, 'steel', 'structural',
        ...     7850, 210000, 81000, 400)
        >>> print(mat1)
        Physical material object
        Name: steel
        uid: 1
        variety: structural
        density: 7850
        E: 210000
        G: 81000
        <BLANKLINE>

    """

    uid: int
    name: str
    variety: str
    density: float
    e_mod: float
    g_mod: float
    f_y: float

    def __srepr__(self):
        """
        Short version of repr.

        """

        return f"Physical material: {self.name}, {self.uid}"

    def __repr__(self):
        res = ""
        res += "Physical material object\n"
        res += f"Name: {self.name}\n"
        res += f"uid: {self.uid}\n"
        res += f"variety: {self.variety}\n"
        res += f"density: {self.density}\n"
        res += f"E: {self.e_mod}\n"
        res += f"G: {self.g_mod}\n"
        return res



"""
Common utility functions used for 3D data visualization
https://plotly.com/python/reference/
"""
#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import plotly.graph_objects as go  # type: ignore
import numpy as np

# pio.renderers.default = 'browser'

# used for production plots
node_marker = {
    'fixed': ("square", 8),
    'release': ("circle-open", 6),
    'free': ("circle", 3),
    'parent': ("circle-open", 15),
    'internal': ("x", 2),
}

# # used for development plots
# node_marker = {
#     "fixed": ("square", 10),
#     "free": ("circle", 5),
#     "parent": ("circle-open", 20),
#     "internal": ("x", 3),
#     "release": ("circle-open", 15),
# }


def global_layout(mdl, camera=None):
    """
    Some general definitions that are often needed.

    """

    # get a proper boudning box form the model
    ref_len = mdl.reference_length()
    p_min, p_max = mdl.bounding_box(padding=2.0 * ref_len)

    # view_type = "orthographic"
    view_type = "perspective"
    if not camera:
        camera = dict(
            up=dict(x=0, y=0, z=1),
            center=dict(x=0, y=0, z=0),
            eye=dict(x=0.7, y=1.00, z=0.30),
            projection={"type": view_type},
        )
    ref_len = np.linalg.norm(p_max - p_min)  # used in aspect ratio calcs
    return go.Layout(
        scene=dict(
            xaxis_visible=False,
            yaxis_visible=False,
            zaxis_visible=False,
            bgcolor="white",
            camera=camera,
            xaxis=dict(range=[p_min[0], p_max[0]], autorange=False),
            yaxis=dict(range=[p_min[1], p_max[1]], autorange=False),
            zaxis=dict(range=[p_min[2], p_max[2]], autorange=False),
            # it's interesting that we have to do this to get the
            # axes aspect ratio right.. but hey! it works
            aspectratio=dict(
                x=(p_max[0] - p_min[0]) / (ref_len / 4.0),
                y=(p_max[1] - p_min[1]) / (ref_len / 4.0),
                z=(p_max[2] - p_min[2]) / (ref_len / 4.0),
            ),
            # note:
            # aspectmode='data': was causing issues with
            # the camera 'moving' across different animation frames
        )
    )


if __name__ == "__main__":
    pass



"""
Library for consistent all-purpose 2D plots
"""


#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

#############
# Resources #
#############
#
# https://coolors.co/  <- amazing color palette generator
#

import sys
import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go

nparr = npt.NDArray[np.float64]


C_LINE = "#6F1D1B"
C_AXIS_LINE = "#432818"
C_TITLE_COLOR = "#432818"
C_LABEL_COLOR = "#432818"
C_GRID = "#BB9457"
C_BACKGROUND = "#FFE6A7"
C_HOVERLABEL_BG = "#BB9457"


def line_plot_interactive(
    title_text,
    x_vals,
    y_vals,
    mode,
    xlab="x",
    xunit=None,
    xhoverformat=".0f",
    ylab="y",
    yunit=None,
    yhoverformat=".0f",
):
    """
    Gneric polty line plot
    """
    assert len(x_vals) == len(y_vals), "Dimensions don't match"
    assert mode in [
        "spline+markers",
        "line",
    ], "mode can either be `spline+markers` or `line`"

    if mode == "line":
        lshape = "linear"
        lmode = "lines"
        lwidth = 2
    elif mode == "spline+markers":
        lshape = "spline"
        lmode = "lines+markers"
        lwidth = 4
    else:
        raise ValueError("oops! This should never run. Strange.")

    if xunit:
        xtitle = f"{xlab} ({xunit})"
    else:
        xtitle = xlab
    if yunit:
        ytitle = f"{ylab} ({yunit})"
    else:
        ytitle = ylab

    num_points = len(x_vals)
    indices: nparr = np.array([range(num_points)])
    my_hovertemplate = (
        "XY value pair: %{customdata[0]:d}<br>"
        + xlab
        + " = %{x: "
        + xhoverformat
        + "} "
    )
    if xunit:
        my_hovertemplate += xunit + "<br>"
    else:
        my_hovertemplate += "<br>"
    my_hovertemplate += ylab + " = %{y:" + yhoverformat + "} "
    if yunit:
        my_hovertemplate += yunit + "<extra></extra>"
    else:
        my_hovertemplate += "<extra></extra>"
    fig = go.Figure(
        data=go.Scatter(
            x=x_vals,
            y=y_vals,
            mode=lmode,
            line_shape=lshape,
            line=dict(color=C_LINE, width=lwidth),
            marker=dict(
                color=C_BACKGROUND, size=10, line=dict(color=C_LINE, width=4)
            ),
            customdata=np.reshape(indices, (-1, 1)),
            hovertemplate=my_hovertemplate,
        )
    )
    fig.update_layout(
        xaxis=dict(
            showline=True,
            showgrid=True,
            gridcolor=C_GRID,
            showticklabels=True,
            linecolor=C_BACKGROUND,
            zerolinecolor=C_AXIS_LINE,
            linewidth=0,
            ticks="outside",
            tickcolor=C_GRID,
            tickfont=dict(
                family="Cambria",
                size=14,
                color=C_AXIS_LINE,
            ),
        ),
        yaxis=dict(
            showline=True,
            showgrid=True,
            gridcolor=C_GRID,
            showticklabels=True,
            linecolor=C_BACKGROUND,
            zerolinecolor=C_AXIS_LINE,
            linewidth=0,
            ticks="outside",
            tickcolor=C_GRID,
            tickfont=dict(
                family="Cambria",
                size=14,
                color=C_AXIS_LINE,
            ),
        ),
        showlegend=False,
        plot_bgcolor=C_BACKGROUND,
    )
    annotations = [
        dict(
            xref="paper",
            yref="paper",
            x=0.0,
            y=1.00,
            xanchor="left",
            yanchor="bottom",
            text=title_text,
            font=dict(family="Cambria", size=28, color=C_TITLE_COLOR),
            showarrow=False,
        )
    ]
    fig.update_layout(annotations=annotations)
    fig.update_layout(
        xaxis_title=xtitle,
        yaxis_title=ytitle,
        font=dict(family="Cambria", size=18, color=C_LABEL_COLOR),
        hoverlabel=dict(
            bgcolor=C_HOVERLABEL_BG, font_size=15, font_family="Cambria"
        ),
    )
    if "pytest" not in sys.modules:
        fig.show()


if __name__ == "__main__":
    pass


# # (Test)
# x = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
# y = [0, 10, 20, 30, 35, 38, 40, 41, 42, 43]

# line_plot_interactive('Pushover\nDirection: Y',
#                       xlab='Displacement', ylab='Base Shear',
#                       xunit='in', yunit='lb', x_vals=x, y_vals=y,
#                       mode='spline+markers')



"""
Defines methods to assign self-weight and self-mass to a loadcase
using a given model.
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import numpy as np
from .. import common
from ..ops import element


def self_weight(mdl, lcase, factor=1.00):
    """
    Assigns the structure's self weight to its members.

    """

    for elm in mdl.list_of_elements():
        if isinstance(
            elm, (element.ElasticBeamColumn, element.DispBeamColumn)
        ):
            weight_per_length = elm.section.weight_per_length()
            # apply weight as UDL
            if elm.visibility.skip_opensees_definition:
                # in that case apply its weight to the connecting nodes
                elm_len = elm.clear_length()
                elm_w = weight_per_length * elm_len * factor
                lcase.node_loads[elm.nodes[0].uid].add(
                    [0.00, 0.00, -elm_w / 2.00, 0.00, 0.00, 0.00]
                )
                lcase.node_loads[elm.nodes[1].uid].add(
                    [0.00, 0.00, -elm_w / 2.00, 0.00, 0.00, 0.00]
                )
            else:
                lcase.line_element_udl[elm.uid].add_glob(
                    np.array([0.0, 0.0, -weight_per_length * factor])
                )
        if isinstance(elm, element.TrussBar):
            weight_per_length = elm.weight_per_length
            # apply its weight to the connecting nodes
            elm_len = elm.clear_length()
            elm_w = weight_per_length * elm_len * factor
            lcase.node_loads[elm.nodes[0].uid].add(
                [0.00, 0.00, -elm_w / 2.00, 0.00, 0.00, 0.00]
            )
            lcase.node_loads[elm.nodes[1].uid].add(
                [0.00, 0.00, -elm_w / 2.00, 0.00, 0.00, 0.00]
            )


def self_mass(mdl, lcase):
    """
    Assigns the structure's self mass to its members

    """

    if mdl.settings.imperial_units:
        g_const = common.G_CONST_IMPERIAL
    else:
        g_const = common.G_CONST_SI

    for elm in mdl.list_of_elements():
        if isinstance(
            elm, (element.ElasticBeamColumn, element.DispBeamColumn)
        ):
            weight_per_length = elm.section.weight_per_length()
        elif isinstance(elm, element.TrussBar):
            weight_per_length = elm.weight_per_length
        else:
            # don't consider other types of elements
            continue

        mass_per_length = weight_per_length / g_const
        # apply lumped mass at the connecting nodes
        half_mass = mass_per_length * elm.clear_length() / 2.00
        lcase.node_mass[elm.nodes[0].uid].add([half_mass] * 3 + [0.00] * 3)
        lcase.node_mass[elm.nodes[1].uid].add([half_mass] * 3 + [0.00] * 3)



"""
Model Generator for OpenSees ~ design library
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
import pandas as pd
from ..solver import Analysis
from .basic_forces import basic_forces
from ..model import Model
from ..solver import ModalResponseSpectrumAnalysis

nparr = npt.NDArray[np.float64]


@dataclass(repr=False)
class LoadCombination:
    """
    Load combinations
    stuff in the first list are added.
    stuff inside the sub-list are enveloped
    """

    mdl: Model
    combo: dict[str, list[tuple[float, Analysis, str]]] = field(
        default_factory=dict
    )

    def envelope_basic_forces(self, elm, num_points):
        """
        Calculates the envelope of the basic forces for the
        given load combination
        """
        df_min = pd.DataFrame(
            np.full((num_points, 6), np.inf),
            columns=["nx", "qy", "qz", "tx", "mz", "my"],
        )
        df_max = pd.DataFrame(
            np.full((num_points, 6), -np.inf),
            columns=["nx", "qy", "qz", "tx", "mz", "my"],
        )
        for component_to_envelope in self.combo.values():
            df_tot = pd.DataFrame(
                np.full((num_points, 6), 0.00),
                columns=["nx", "qy", "qz", "tx", "mz", "my"],
            )
            for component_to_add in component_to_envelope:
                factor, anl, case_name_str = component_to_add
                res = basic_forces(anl, case_name_str, 0, elm, num_points)
                assert isinstance(res, pd.DataFrame)
                dframe = res * factor
                df_tot += dframe
            df_min[df_min > df_tot] = df_tot
            df_max[df_max < df_tot] = df_tot

        return df_min, df_max

    def envelope_node_displacement(self, node):
        """
        Calculates the enveloped node displacement
        """
        disp_min: nparr = np.full(6, np.inf)
        disp_max: nparr = np.full(6, -np.inf)
        for component_to_envelope in self.combo.values():
            disp_tot = np.full(6, 0.00)
            for component_to_add in component_to_envelope:
                factor, anl, case_name_str = component_to_add
                if isinstance(anl, ModalResponseSpectrumAnalysis):
                    disp: nparr = np.array(
                        anl.combined_node_disp(node.uid) * factor
                    )
                else:
                    disp = np.array(
                        anl.results[case_name_str].node_displacements[
                            node.uid
                        ][0]
                    )
                disp_tot += disp
            disp_min[disp_min > disp_tot] = disp_tot[disp_min > disp_tot]
            disp_max[disp_max < disp_tot] = disp_tot[disp_max < disp_tot]

        return disp_min, disp_max

    def envelope_node_displacement_diff(self, node_i, node_j):
        """
        Calculates the enveloped displacement difference between
        two nodes
        """
        disp_min = np.full(6, np.inf)
        disp_max = np.full(6, -np.inf)
        for component_to_envelope in self.combo.values():
            disp_tot = np.full(6, 0.00)
            for component_to_add in component_to_envelope:
                factor, anl, case_name_str = component_to_add
                if isinstance(anl, ModalResponseSpectrumAnalysis):
                    disp: nparr = np.array(
                        anl.combined_node_disp_diff(node_i.uid, node_j.uid)
                        * factor
                    )
                else:
                    disp_i: nparr = np.array(
                        anl.results[case_name_str].node_displacements[
                            node_i.uid
                        ][0]
                    )
                    disp_j: nparr = np.array(
                        anl.results[case_name_str].node_displacements[
                            node_j.uid
                        ][0]
                    )
                    disp = disp_i - disp_j
                disp_tot += disp
            disp_min[disp_min > disp_tot] = disp_tot[disp_min > disp_tot]
            disp_max[disp_max < disp_tot] = disp_tot[disp_max < disp_tot]

        return disp_min, disp_max



"""
Determine the latest version of the osmg package on PyPI
"""

import argparse
import requests


def get_latest_version(package_name):
    """
    Determine the latest version of the osmg package on PyPI
    """
    url = f"https://pypi.org/pypi/{package_name}/json"

    response = requests.get(url, timeout=10)
    package_info = response.json()
    latest_version = package_info["info"]["version"]
    return latest_version


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--package_name", default="osmg", help="Name of the package."
    )
    args = parser.parse_args()

    package = args.package_name
    version = get_latest_version(package)
    print(version)



"""
Objects that introduce nodes to a model.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from dataclasses import dataclass
from ..ops.node import Node

if TYPE_CHECKING:
    from .model import Model


@dataclass(repr=False)
class NodeGenerator:
    """
    NodeGenerator objects introduce node objects to a model.

    Attributes:
      model: Model to act upon.

    """

    model: Model

    def add_node_lvl(self, x_loc: float, y_loc: float, lvl: int) -> Node:
        """
        Adds a node at the specified coordinates and level.

        Arguments:
          x_loc: x coordinate of the node.
          y_loc: y coordinate of the node.
          lvl: Level at which the node should be added.

        Returns:
          The node object that was added.

        Example:
          >>> from osmg.model import Model
          >>> from osmg.gen.node_gen import NodeGenerator
          >>> model = Model('test_model')
          >>> model.add_level(0, 0.00)
          >>> generator = NodeGenerator(model)
          >>> generator.add_node_lvl(2.00, 3.00, 0)
          Node object
            uid: 0
            coords: [2.0, 3.0, 0.0]
            restraint: [False, False, False, False, False, False]
          <BLANKLINE>

        """
        lvls = self.model.levels
        level = lvls[lvl]
        node = Node(
            uid=self.model.uid_generator.new("node"),
            coords=[x_loc, y_loc, level.elevation],
        )
        level.nodes.add(node)
        return node

    def add_node_lvl_xyz(
        self, x_loc: float, y_loc: float, z_loc: float, lvl: int
    ) -> Node:
        """
        Adds a node at the specified coordinates and level having a
        custom elevation.

        Arguments:
            x_loc: x coordinate of the node.
            y_loc: y coordinate of the node.
            z_loc: Elevation of the node.
            lvl: Level uid at which the node should be added.

        Returns:
            The node object that was added.

        Example:
            >>> from osmg.model import Model
            >>> from osmg.gen.node_gen import NodeGenerator
            >>> model = Model('test_model')
            >>> model.add_level(0, 0.00)
            >>> generator = NodeGenerator(model)
            >>> generator.add_node_lvl_xyz(2.00, 3.00, 4.00, 0)
            Node object
              uid: 0
              coords: [2.0, 3.0, 4.0]
              restraint: [False, False, False, False, False, False]
            <BLANKLINE>

        """
        lvls = self.model.levels
        level = lvls[lvl]
        node = Node(
            uid=self.model.uid_generator.new("node"),
            coords=[x_loc, y_loc, z_loc],
        )
        level.nodes.add(node)
        return node

    def add_node_active(self, x_loc, y_loc):
        """
        Adds a node[/s] at the specified coordinates to all active levels.

        Arguments:
            x_loc: x coordinate of the node[/s].
            y_loc: y coordinate of the node[/s].

        Example:
            >>> from osmg.model import Model
            >>> from osmg.gen.node_gen import NodeGenerator
            >>> model = Model('test_model')
            >>> model.add_level(0, 0.00)
            >>> model.add_level(1, 1.00)
            >>> model.add_level(2, 2.00)
            >>> model.levels.set_active([1, 2])
            >>> generator = NodeGenerator(model)
            >>> generator.add_node_active(2.00, 3.00)
            >>> model.levels[0].nodes.__srepr__()
            '[Collection of 0 items]'
            >>> model.levels[1].nodes.__srepr__()
            '[Collection of 1 items]'
            >>> model.levels[2].nodes.__srepr__()
            '[Collection of 1 items]'

        """
        lvls = self.model.levels
        assert lvls.active, "No active levels."
        for key in lvls.active:
            self.add_node_lvl(x_loc, y_loc, key)



"""
Defines :obj:`~osmg.ops.node.Node` objects.
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from dataclasses import dataclass, field
from functools import total_ordering
from ..graphics.visibility import NodeVisibility


@dataclass
@total_ordering
class Node:
    """
    OpenSees node
    https://openseespydoc.readthedocs.io/en/latest/src/node.html?highlight=node

    Attributes:
        uid: Unique ID of the node.
        coords: List of node coordinates.
        restraint: List of boolean values identifying whether the
          corresponding DOF is restrained.

    """

    uid: int
    coords: list[float]
    restraint: list[bool] = field(init=False)
    visibility: NodeVisibility = field(default_factory=NodeVisibility)

    def __post_init__(self):
        self.restraint = [False] * 6

    def __le__(self, other):
        return self.uid < other.uid

    def __repr__(self):
        res = ""
        res += "Node object\n"
        res += f"  uid: {self.uid}\n"
        res += f"  coords: {self.coords}\n"
        res += f"  restraint: {self.restraint}\n"
        return res



"""
Introduces some default objects to a model.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
import numpy.typing as npt
from .physical_material import PhysicalMaterial
from . import common
from .ops.uniaxial_material import Elastic
from .ops.uniaxial_material import Steel02
from .ops.section import ElasticSection
from .gen.section_gen import SectionGenerator


if TYPE_CHECKING:
    from .model import Model

nparr = npt.NDArray[np.float64]


def load_util_rigid_elastic(model: Model) -> None:
    """
    Adds a default rigid elastic beamcolumn element
    to the model.

    """

    new_uid = model.uid_generator.new("section")
    sec = ElasticSection(
        name="rigid_link_section",
        uid=new_uid,
        outside_shape=None,
        snap_points=None,
        e_mod=common.STIFF,
        area=1.00,
        i_y=1.00,
        i_x=1.00,
        g_mod=common.STIFF,
        j_mod=1.00,
        sec_w=0.00,
    )
    model.elastic_sections.add(sec)


def load_default_elastic(model: Model, sec_name: str) -> None:
    """
    Adds default non-rigid elastic beamcolumn element
    sections to the model.

    """

    # intantiate a section generator object for the model
    sgen = SectionGenerator(model)
    # generate a default elastic section and add it to the model
    sgen.generate_generic_elastic(
        name=sec_name, e_times_a=1.00, e_times_i=1.00, g_times_j=1.00
    )


def load_default_steel(model: Model) -> None:
    """
    Adds a default steel material to the model.

    Note: If different properties are required, the values provided
    here can be altered, or the materials can be defined at the
    user-side and added to the model instead of calling this method.

    """

    if model.settings.imperial_units:
        # force: lb, length: in
        uniaxial_mat = Steel02(
            model.uid_generator.new("uniaxial material"),
            "default steel",
            55000.00,
            29000000.00,
            11153846.15,
            0.01,
            15.0,
            0.925,
            0.15,
        )
        physical_mat = PhysicalMaterial(
            model.uid_generator.new("physical material"),
            "default steel",
            "A992-Fy50",
            0.2835648148148148 / common.G_CONST_IMPERIAL,
            29000000.00,
            11153846.15,
            55000.00,
        )
    else:
        # force: kN, length: m, but it's just the values above,
        # converted to SI, instead of the properties of 'typical'
        # european steel.
        uniaxial_mat = Steel02(
            model.uid_generator.new("uniaxial material"),
            "default steel",
            379211.65,
            199947961.12,
            76903062.09,
            0.01,
            15.0,
            0.925,
            0.15,
        )
        physical_mat = PhysicalMaterial(
            model.uid_generator.new("physical material"),
            "default steel",
            "A992-Fy50",
            773.37062406 / common.G_CONST_IMPERIAL,
            199947961.12,
            76903062.09,
            379211.65,
        )
    model.uniaxial_materials.add(uniaxial_mat)
    model.physical_materials.add(physical_mat)


def load_default_fix_release(model: Model) -> None:
    """
    Loads default fix and release elastic uniaxial materials
    used to simulate moment releases using zerolength elements.

    """

    uniaxial_mat = Elastic(
        uid=model.uid_generator.new("uniaxial material"),
        name="fix_rot",
        e_mod=common.STIFF_ROT,
    )
    model.uniaxial_materials.add(uniaxial_mat)
    uniaxial_mat = Elastic(
        uid=model.uid_generator.new("uniaxial material"),
        name="fix",
        e_mod=common.STIFF,
    )
    model.uniaxial_materials.add(uniaxial_mat)
    uniaxial_mat = Elastic(
        uid=model.uid_generator.new("uniaxial material"),
        name="release",
        e_mod=common.TINY,
    )
    model.uniaxial_materials.add(uniaxial_mat)



"""
Collections are designated containers of objects of a particular type.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import Any
from typing import List
from typing import TypeVar
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from . import common
from .ops import node
from .ops import element


nparr = npt.NDArray[np.float64]

# pylint: disable=invalid-name
TK = TypeVar("TK")
TV = TypeVar("TV")
# pylint: enable=invalid-name


@dataclass(repr=False)
class Collection(dict[TK, TV]):
    """
    Collection of objects.

    Attributes:
        parent: Object to which the Collection belongs.

    Example:
        >>> # obj_collections require parent objects to which they belong
        >>> parent = 52
        >>> my_collection = Collection(parent=parent)
        >>> my_collection.parent
        52
        >>> # now `my_collection` knows that it belongs to the object `52`.
        >>>
        >>>
        >>>
        >>> # Collection.add method
        >>>
        >>> from osmg.ops.node import Node
        >>> my_collection = Collection(parent=None)
        >>> len(my_collection)
        0
        >>> # something to add:
        >>> new_node = Node(uid=0, coords=[0.00, 0.00])
        >>> # add the node to the collection
        >>> my_collection.add(new_node)
        >>> # it has been added:
        >>> len(my_collection)
        1
        >>> # and it can be accessed by its uid:
        >>> id(new_node) == id(my_collection[0])
        True
        >>> # adding something without a uid fails:
        >>> my_collection.add(42)
        Traceback (most recent call last):
            ...
        KeyError: 'Object does not have a uid attribute'
        >>> # adding an object with the same uid fails:
        >>> my_collection.add(new_node)
        Traceback (most recent call last):
            ...
        KeyError: 'uid 0 already exists'
        >>>
        >>>
        >>>
        >>> # obj_collections.retrieve_by_attr method:
        >>>
        >>> from osmg.ops.section import Section
        >>> sec_collection = Collection(parent=None)
        >>> sec_collection.add(Section(name='sec_1', uid=0))
        >>> sec_collection.add(Section(name='sec_2', uid=1))
        >>> sec_collection.retrieve_by_attr('name', 'sec_1')
        Section(name='sec_1', uid=0)
        >>> # we 'll use this later:
        >>>
        >>>
        >>>
        >>> # obj_collections.__srepr__ method:
        >>>
        >>> sec_collection.__srepr__()
        '[Collection of 2 items]'

    """

    parent: Any = field(repr=False)

    def add(self, obj: Any) -> None:
        """
        Add an object to the collection. The object needs to have
        a unique id attribute, `uid`.

        Arguments:
          obj: Object to be added.

        """
        if not hasattr(obj, "uid"):
            raise KeyError("Object does not have a uid attribute")
        if obj.uid in self:
            raise KeyError(f"uid {obj.uid} already exists")
        self[obj.uid] = obj

    def retrieve_by_attr(self, attr: str, val: Any) -> Any:
        """
        Retrieve an object from the collection based on an attribute
        value. If more than one instances satisfy the criterion, the
        function returns the first occurrence.

        Arguments:
          attr: The name of the attribute.
          val: The value that the attribute should have.

        """
        res = None
        for thing in self.values():
            found = False
            if hasattr(thing, attr):
                other_val = getattr(thing, attr)
                if other_val == val:
                    res = thing
                    found = True
                    break
            if found:
                break
        if res is None:
            raise ValueError(
                f'Item having the value "{val}" '
                f'in the attribute "{attr}" not found in collection.'
            )
        return res

    def __srepr__(self):
        """
        Concise version of `repr`.

        """
        return f"[Collection of {len(self)} items]"

    def __repr__(self):
        res = ""
        res += "Collection Object\n"
        res += f"ID: {id(self)}\n"
        res += f"Parent object: {self.parent}\n"
        res += f"Registry size: {len(self)}\n"
        return res


@dataclass(repr=False)
class CollectionActive(Collection[TK, TV]):
    """
    Collection with support for currently active objects.

    Attributes:
      active: List of unique IDs that correspond to the active
        objects.

    Example:
        >>> from osmg.level import Level
        >>> my_collection = CollectionActive(parent=None)
        >>> my_collection.add(Level(
        ...     parent_model=None,
        ...     uid=0,
        ...     elevation=0.00))
        >>> my_collection.add(Level(
        ...     parent_model=None,
        ...     uid=1,
        ...     elevation=1.00))
        >>>
        >>>
        >>>
        >>> # CollectionActive.set_active method:
        >>>
        >>> # set them both as active
        >>> my_collection.set_active([0, 1])
        >>> my_collection.active
        [0, 1]
        >>> # if uid is not present, it fails
        >>> my_collection.set_active([2])
        Traceback (most recent call last):
            ....
        KeyError: 'uid 2 not present in collection.'

    """

    active: List[TK] = field(default_factory=list)

    def set_active(self, uids: List[TK]) -> None:
        """
        Sets the active objects.

        Arguments:
            uids: uids of the objects to set as active

        """
        for uid in uids:
            if uid not in self:
                raise KeyError(f"uid {uid} not present in collection.")
        self.active = uids

    def set_active_all(self):
        """
        Sets the active objects.

        Arguments:
            uids: uids of the objects to set as active

        """
        self.active = []
        for key in self:
            self.active.append(key)


@dataclass(repr=False)
class NodeCollection(Collection[int, node.Node]):
    """
    Node collection.

    Attributes:
        parent: Object to which the Collection belongs.

    Example:
        >>> from osmg.ops.node import Node
        >>> from osmg.level import Level
        >>> level = Level(parent_model=None, uid=0, elevation=0.00)
        >>>
        >>> my_collection = NodeCollection(parent=level)
        >>>
        >>> n_1 = Node(uid=0, coords=[0.00, 0.00, 0.00])
        >>> n_2 = Node(uid=1, coords=[1.00, 1.00, 0.00])
        >>> my_collection.add(n_1)
        >>> my_collection.add(n_2)
        >>>
        >>>
        >>>
        >>> # NodeCollection.search_xy method:
        >>>
        >>> retrieved_node = my_collection.search_xy(1.00, 1.00)
        >>> id(retrieved_node) == id(n_2)  # should be the same object
        True

    """

    named_contents: dict[str, node.Node] = field(default_factory=dict)

    def search_xy(self, x_coord, y_coord):
        """
        Returns the node that occupies a given point if it exists
        """

        candidate_pt: nparr = np.array(
            [x_coord, y_coord, self.parent.elevation]
        )
        for other_node in self.values():
            other_pt: nparr = np.array(other_node.coords)
            if np.linalg.norm(candidate_pt - other_pt) < common.EPSILON:
                return other_node
        # indent the following line once to the right, and you'll be
        # spending a couple of nights trying to figure out why there
        # are so many duplicate nodes
        return None


@dataclass(repr=False)
class CollectionWithConnectivity(Collection[TK, TV]):
    """
    Collection of elements for which it is important to consider their
    connectivity.

    Attributes:
        parent: Object to which the Collection belongs.

    """

    named_contents: dict[str, element.ElasticBeamColumn] = field(
        default_factory=dict
    )

    def add(self, obj):
        """
        Adds an element to the collection.
        The method also checks to see if an object having the same
        connectivity exists in the collection, and raises an error if
        it does.

        Arguments:
          obj: Object to be added.

        """
        uids = [nd.uid for nd in obj.nodes]
        uids.sort()
        uids_tuple = (*uids,)
        if uids_tuple in obj.parent_component.element_connectivity():
            raise ValueError("This should never happen!")
        super().add(obj)



"""
Objects that generate unique IDs

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from dataclasses import dataclass
from itertools import count


@dataclass
class UIDGenerator:
    """
    Generates unique identifiers (uids) for various objects.

    """

    def new(self, thing: str) -> int:
        """
        Generates a new uid for an object of the given type.

        Arguments:
            object_type: The type of object for which to generate a uid.

        Returns:
            A unique identifier for an object of the given type.

        Example:
            >>> from osmg.gen.uid_gen import UIDGenerator
            >>> generator = UIDGenerator()
            >>> generator.new('node')
            0
            >>> generator.new('node')
            1
            >>> generator.new('element')
            0
            >>> generator.new('element')
            1

        """
        if hasattr(self, thing):
            res = next(getattr(self, thing))
            assert isinstance(res, int)
        else:
            setattr(self, thing, count(0))
            res = next(getattr(self, thing))
            assert isinstance(res, int)
        return res



"""
BRB element generator.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from ...gen.mesh_shapes import rect_mesh
from ...model import Model
from ...gen.component_gen import TrussBarGenerator
from ...ops.uniaxial_material import Steel4
from ...ops.uniaxial_material import Fatigue


# pylint: disable=invalid-name

nparr = npt.NDArray[np.float64]


@dataclass(repr=False)
class BRBGenSettings:
    """
    This object holds default values that are used by BRBGenerator
    objects, to better organize those values and improve the usability
    of the BRBGenerator objects by allowing their methods to have
    fewer arguments.

    """

    steel4_b_k: float = field(default=0.003)
    steel4_b_kc: float = field(default=0.023)
    steel4_R_0: float = field(default=25.00)
    steel4_R_0c: float = field(default=25.00)
    steel4_r_1: float = field(default=0.90)
    steel4_r_1c: float = field(default=0.90)
    steel4_r_2: float = field(default=0.15)
    steel4_r_2c: float = field(default=0.15)
    steel4_b_i: float = field(default=0.0025)
    steel4_b_ic: float = field(default=0.0045)
    steel4_rho_i: float = field(default=1.34)
    steel4_rho_ic: float = field(default=0.77)
    steel4_b_l: float = field(default=0.004)
    steel4_b_lc: float = field(default=0.004)
    steel4_R_i: float = field(default=1.0)
    steel4_R_ic: float = field(default=1.0)
    steel4_l_yp: float = field(default=1.0)
    fatigue_e_mod: float = field(default=0.12)
    fatigue_var_m: float = field(default=-0.458)
    transf_type: str = field(default='Corotational')


@dataclass(repr=False)
class BRBGenerator:
    """
    Uses the lower-level `gen` classes to simplify the definition of
    BRB elements simulated using a corotational truss opensees element
    and the Steel4 material wrapped under a Fatigue material, as done
    in [1] and [2].

    1. Zsarnoczay, Adam. "Experimental and numerical investigation of
    buckling restrained braced frames for Eurocode conform design
    procedure development." (2013).

    2. Simpson, Barbara Gwynne. Design development for steel
    strongback braced frames to mitigate concentrations of
    damage. University of California, Berkeley, 2018.

    """

    model: Model
    settings: BRBGenSettings = field(default_factory=BRBGenSettings)

    def add_brb(
        self,
        xi_coord: float,
        yi_coord: float,
        lvl_key_i: int,
        offset_i: nparr,
        snap_i: str,
        xj_coord: float,
        yj_coord: float,
        lvl_key_j: int,
        offset_j: nparr,
        snap_j: str,
        area: float,
        f_y: float,
        e_0: float,
        casing_size: float,
        unit_weight: float,
    ) -> None:
        """
        Adds a BRB element to the model.

        Parameters:
          xi_coord: x coordinate of the i-end
          yi_coord: y coordinate of the i-end
          lvl_key_i: Level ID of the i-end
          offset_i: Rigid offset at the i-end
          snap_i: Snap tag for the i-end connectivity used to
            automatically determine an offset based on existing
            elements.
          xj_coord: x coordinate of the j-end
          yj_coord: y coordinate of the j-end
          lvl_key_j: Level ID of the j-end
          offset_j: Rigid offset at the j-end
          snap_j: Snap tag for the j-end connectivity used to
            automatically determine an offset based on existing
            elements.
          area: Core area of the BRB element.
          f_y: Steel stress.
          e_0: Young's modulus, adjusted to account for
            non-prismatic shape. See [2], Figure 4.15, and Equations
            4.7, 4.8.
          casing_size: Assuming a square cross-section,
            controls the side length. Used for visualization purposes
            and to determine the weight of the BRB element, which is
            lumped to the connecting nodes. This parameter has no
            effect on the force-deformation relationship of the
            element.
          unit_weight: Weight per unit volume, used together
            with the casing_size, assuming a square section and a
            workpoint-to-workpoint length, in order to determine the
            weight of the BRB, which is lumped to the two connecting
            nodes of the Truss element as weight and mass when the
            self-weight and self-mass preprocessing methods are
            called.

        """

        trg = TrussBarGenerator(self.model)

        uid = self.model.uid_generator.new("uniaxial material")
        mat = Steel4(
            uid=uid,
            name=f'auto_BRB_{uid}',
            Fy=f_y,
            E0=e_0,
            b_k=self.settings.steel4_b_k,
            b_kc=self.settings.steel4_b_kc,
            R_0=self.settings.steel4_R_0,
            R_0c=self.settings.steel4_R_0c,
            r_1=self.settings.steel4_r_1,
            r_1c=self.settings.steel4_r_1c,
            r_2=self.settings.steel4_r_2,
            r_2c=self.settings.steel4_r_2c,
            b_i=self.settings.steel4_b_i,
            b_ic=self.settings.steel4_b_ic,
            rho_i=self.settings.steel4_rho_i,
            rho_ic=self.settings.steel4_rho_ic,
            b_l=self.settings.steel4_b_l,
            b_lc=self.settings.steel4_b_lc,
            R_i=self.settings.steel4_R_i,
            R_ic=self.settings.steel4_R_ic,
            l_yp=self.settings.steel4_l_yp,
        )
        uid = self.model.uid_generator.new("uniaxial material")
        mat_fatigue = Fatigue(
            uid=uid,
            name=f'auto_BRB_fatigue_{uid}',
            predecessor=mat,
            e_mod=self.settings.fatigue_e_mod,
            var_m=self.settings.fatigue_var_m,
        )
        trg.add(
            xi_coord,
            yi_coord,
            lvl_key_i,
            offset_i,
            snap_i,
            xj_coord,
            yj_coord,
            lvl_key_j,
            offset_j,
            snap_j,
            transf_type=self.settings.transf_type,
            area=area,
            mat=mat_fatigue,
            outside_shape=rect_mesh(casing_size, casing_size),
            weight_per_length=unit_weight * casing_size**2,
        )



"""
Objects used to retrieve other objects, or information related to
them, based on certain criteria.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

# pylint: disable=inconsistent-return-statements

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Union
from typing import Optional
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
from ..line import Line
from ..load_case import LoadCase
from .. import common
from ..ops import element

if TYPE_CHECKING:
    from ..component_assembly import ComponentAssembly
    from ..ops.node import Node
    from ..model import Model


nparr = npt.NDArray[np.float64]


@dataclass(repr=False)
class ElmQuery:
    """
    Used to retrieve nodes and component
    assemblies.

    Attributes:
      model: Model to be searched.

    """

    model: Model

    def search_connectivity(
        self, nodes: list[Node]
    ) -> Optional[ComponentAssembly]:
        """
        Finds component assembly objects based on the nodes the are
        connected to.

        Arguments:
          nodes: List containing nodes. If a component assembly is
            connected to those nodes, it is returned, otherwise the
            method returns None. If the component assembly is
            partially connected to the nodes (e.g. connected to them,
            but also to other nodes), it is not returned.

        """
        uids = [node.uid for node in nodes]
        uids.sort()
        uids_tuple = (*uids,)
        conn_dict = self.model.component_connectivity()
        val = conn_dict.get(uids_tuple)
        return val

    def search_node_lvl(
        self,
        x_loc: float,
        y_loc: float,
        lvl: int,
        z_loc: Optional[float] = None,
        internal: bool = False,
    ) -> Optional[Node]:
        """
        Looks if a primary node exists at the specified location.

        Arguments:
          x_loc: x-coordinate
          y_loc: y-coordinate
          lvl: Key of the level to be searched.

        """
        lvls = self.model.levels
        level = lvls[lvl]
        res = None
        # check to see if node exists
        if z_loc:
            candidate_pt: nparr = np.array([x_loc, y_loc, z_loc])
            ndims = 3
        else:
            candidate_pt = np.array([x_loc, y_loc])
            ndims = 2
        nodes = level.nodes
        if internal:
            for comp in level.components.values():
                nodes.update(comp.internal_nodes)
        for other_node in nodes.values():
            other_pt: nparr = np.array(other_node.coords[:ndims])
            if np.linalg.norm(candidate_pt - other_pt) < common.EPSILON:
                res = other_node
                break
        return res

    def retrieve_components_from_nodes(
        self, nodes: list[Node], lvl_uid: Optional[int] = None
    ) -> dict[int, ComponentAssembly]:
        """
        Retrieves component assemblies if at least one of their
        external nodes matches the given list of nodes.

        Arguments:
          nodes: List of primary nodes.
          lvl_uid: ID of the level to be searched. If None is
            specified, the method searches all levels.

        """
        retrieved_components = {}
        if lvl_uid:
            level = self.model.levels[lvl_uid]
            candidate_components = level.components.values()
        else:
            candidate_components = self.model.list_of_components()
        given_node_uids = [n.uid for n in nodes]
        for component in candidate_components:
            accept = False
            external_nodes = component.external_nodes.values()
            for node in external_nodes:
                if node.uid in given_node_uids:
                    accept = True
                    continue
            if accept:
                retrieved_components[component.uid] = component
        return retrieved_components

    def retrieve_component_from_nodes(
        self, nodes: list[Node], lvl_uid: Optional[int] = None
    ) -> Optional[ComponentAssembly]:
        """
        Retrieves a single component assembly if all of its external
        nodes match the given list of nodes.

        Arguments:
          nodes: List of primary nodes.
          lvl_uid: Key of the level to be searched. If None is
            specified, the method searches all levels.

        """
        retrieved_component = None
        if lvl_uid:
            level = self.model.levels[lvl_uid]
            candidate_components = level.components.values()
        else:
            candidate_components = self.model.list_of_components()
        given_node_uids = [n.uid for n in nodes]
        for component in candidate_components:
            reject = False
            external_nodes = component.external_nodes.values()
            for node in external_nodes:
                if node.uid not in given_node_uids:
                    reject = True
                    continue
            if not reject:
                retrieved_component = component
        return retrieved_component

    def retrieve_component(
        self, x_loc: float, y_loc: float, lvl: int
    ) -> Optional[ComponentAssembly]:
        """
        Retrieves a component assembly of a level if any of its
        line elements passes trhough the specified point.

        Arguments:
          x_loc: x-coordinate
          y_loc: y-coordinate
          lvl: Key of the level to be searched.

        Returns:
          The first element found, None otherwise.

        """
        level = self.model.levels[lvl]
        for component in level.components.values():
            if len(component.external_nodes) != 2:
                continue
            line_elems: list[
                Union[
                    element.TrussBar,
                    element.ElasticBeamColumn,
                    element.DispBeamColumn,
                ]
            ] = []
            for elm in component.elements.values():
                if isinstance(elm, element.TrussBar):
                    line_elems.append(elm)
                if isinstance(elm, element.ElasticBeamColumn):
                    line_elems.append(elm)
                if isinstance(elm, element.DispBeamColumn):
                    line_elems.append(elm)

            for elm in line_elems:
                if isinstance(elm, element.TrussBar):
                    p_i = np.array(elm.nodes[0].coords)
                    p_j = np.array(elm.nodes[1].coords)
                else:
                    p_i = (
                        np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
                    )
                    p_j = (
                        np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
                    )
                if np.linalg.norm(p_i[0:2] - p_j[0:2]) < common.EPSILON:
                    if (
                        np.linalg.norm(np.array((x_loc, y_loc)) - p_i[0:2])
                        < common.EPSILON
                    ):
                        return component
                else:
                    line = Line("", p_i[0:2], p_j[0:2])
                    line.intersects_pt(np.array((x_loc, y_loc)))
                    if line.intersects_pt(np.array((x_loc, y_loc))):
                        return component
        return None


@dataclass
class LoadCaseQuery:
    """
    Load case query objects retrieve information associated with load
    cases.

    Attributes:
      model: Model object to be searched.
      loadcase: LoadCase object to be searched.

    """

    model: Model
    loadcase: LoadCase

    def level_masses(self):
        """
        Returns the total mass of each level.

        """
        mdl = self.model
        num_lvls = len(mdl.levels)
        distr = np.zeros(num_lvls)
        for key, lvl in mdl.levels.items():
            for node in lvl.nodes.values():
                mass = self.loadcase.node_mass[node.uid]
                distr[key] += mass.val[0]

            for component in lvl.components.values():
                for node in component.internal_nodes.values():
                    mass = self.loadcase.node_mass[node.uid]
                    distr[key] += mass.val[0]
        for uid, node in self.loadcase.parent_nodes.items():
            distr[uid] += self.loadcase.node_mass[node.uid].val[0]
        return distr



"""
Defines OpenSees Element interfrace objects.
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from dataclasses import dataclass, field
from typing import Union
from typing import Optional
from typing import TYPE_CHECKING
import numpy as np
import numpy.typing as npt
from .uniaxial_material import UniaxialMaterial
from .node import Node
from .section import ElasticSection
from .section import FiberSection
from ..mesh import Mesh
from ..graphics.visibility import ElementVisibility

if TYPE_CHECKING:
    from ..component_assembly import ComponentAssembly


nparr = npt.NDArray[np.float64]


@dataclass(repr=False)
class Element:
    """
    OpenSees element
    https://openseespydoc.readthedocs.io/en/latest/src/element.html

    Attributes:
        parent_component:
          the parent component assembly that this element belongs to.
        uid: the unique identifier of this element.
        nodes: the list of nodes that this element connects.

    """

    parent_component: ComponentAssembly = field(repr=False)
    uid: int
    nodes: list[Node]
    visibility: ElementVisibility = field(init=False)

    def __post_init__(self):
        self.visibility = ElementVisibility()


@dataclass(repr=False)
class ZeroLength(Element):
    """
    OpenSees ZeroLength element
    https://openseespydoc.readthedocs.io/en/latest/src/ZeroLength.html

    """

    mats: list[UniaxialMaterial]
    dirs: list[int]
    vecx: nparr
    vecyp: nparr

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return [
            "zeroLength",
            self.uid,
            *[n.uid for n in self.nodes],
            "-mat",
            *[m.uid for m in self.mats],
            "-dir",
            *self.dirs,
            "-doRayleigh",
            1,
            "-orient",
            *self.vecx,
            *self.vecyp,
        ]

    def __repr__(self):
        res = ""
        res += "ZeroLength element object\n"
        res += f"uid: {self.uid}"
        res += "Materials:"
        for mat, direction in zip(self.mats, self.dirs):
            res += f"  {direction}: {mat.name}\n"
        res += f"vecx: {self.vecx}\n"
        res += f"vecyp: {self.vecyp}\n"
        return res


@dataclass(repr=False)
class TwoNodeLink(Element):
    """
    OpenSees TwoNodeLink element
    https://openseespydoc.readthedocs.io/en/latest/src/twoNodeLink.html

    """

    mats: list[UniaxialMaterial]
    dirs: list[int]
    vecx: nparr
    vecyp: nparr

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return [
            "twoNodeLink",
            self.uid,
            *[n.uid for n in self.nodes],
            "-mat",
            *[m.uid for m in self.mats],
            "-dir",
            *self.dirs,
            "-orient",
            *self.vecyp,
        ]

    def __repr__(self):
        res = ""
        res += "TwoNodeLink element object\n"
        res += f"uid: {self.uid}"
        res += "Materials:"
        for mat, direction in zip(self.mats, self.dirs):
            res += f"  {direction}: {mat.name}\n"
        res += f"vecx: {self.vecx}\n"
        res += f"vecyp: {self.vecyp}\n"
        return res


@dataclass(repr=False)
class TrussBar(Element):
    """
    Implements both of the following:
    OpenSees Truss Element
    https://openseespydoc.readthedocs.io/en/latest/src/trussEle.html
    OpenSees Corotational Truss Element
    https://openseespydoc.readthedocs.io/en/latest/src/corotTruss.html

    """

    transf_type: str
    area: float
    mat: UniaxialMaterial
    outside_shape: Optional[Mesh] = field(default=None)
    weight_per_length: float = field(default=0.00)
    rho: float = field(default=0.00)
    cflag: int = field(default=0)
    rflag: int = field(default=0)

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        elm_name = {'Linear': 'Truss', 'Corotational': 'corotTruss'}

        return [
            elm_name[self.transf_type],
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.area,
            self.mat.uid,
            '-rho',
            self.rho,
            '-cMass',
            self.cflag,
            '-doRayleigh',
            self.rflag,
        ]

    def clear_length(self):
        """
        Returns the clear length of the element (without the rigid
        offsets)

        """

        p_i = np.array(self.nodes[0].coords)
        p_j = np.array(self.nodes[1].coords)
        return np.linalg.norm(p_i - p_j)

    def __repr__(self):
        elm_name = {'Linear': 'Truss', 'Corotational': 'corotTruss'}
        res = ""
        res += f"{elm_name[self.transf_type]} element object\n"
        res += f"uid: {self.uid}\n"
        res += f"node_i.uid: {self.nodes[0].uid}\n"
        res += f"node_j.uid: {self.nodes[1].uid}\n"
        res += f"node_i.coords: {self.nodes[0].coords}\n"
        res += f"node_j.coords: {self.nodes[1].coords}\n"
        return res


@dataclass(repr=False)
class GeomTransf:
    """
    OpenSees geomTransf object
    https://openseespydoc.readthedocs.io/en/latest/src/ZeroLength.html
    https://openseespydoc.readthedocs.io/en/latest/src/geomTransf.html?highlight=geometric%20transformation

    """

    transf_type: str
    uid: int
    offset_i: nparr
    offset_j: nparr
    x_axis: nparr
    y_axis: nparr
    z_axis: nparr

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return [
            self.transf_type,
            self.uid,
            *self.z_axis,
            "-jntOffset",
            *self.offset_i,
            *self.offset_j,
        ]


@dataclass(repr=False)
class ElasticBeamColumn(Element):
    """
    OpenSees Elastic Beam Column Element
    https://openseespydoc.readthedocs.io/en/latest/src/elasticBeamColumn.html

    """

    section: ElasticSection
    geomtransf: GeomTransf
    n_x: Optional[float] = field(default=None)
    n_y: Optional[float] = field(default=None)

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        if self.n_x is not None:
            n_x = self.n_x
            k44_x = 6.0 * (1.0 + n_x) / (2.0 + 3.0 * n_x)
            k11_x = (1.0 + 2.0 * n_x) * k44_x / (1.0 + n_x)
            k33_x = k11_x

        if self.n_y is not None:
            n_y = self.n_y
            k44_y = 6.0 * (1.0 + n_y) / (2.0 + 3.0 * n_y)
            k11_y = (1.0 + 2.0 * n_y) * k44_y / (1.0 + n_y)
            k33_y = k11_y

        if self.n_x is not None and self.n_y is None:
            return [
                "ModElasticBeam3d",
                self.uid,
                self.nodes[0].uid,
                self.nodes[1].uid,
                self.section.area,
                self.section.e_mod,
                self.section.g_mod,
                self.section.j_mod,
                self.section.i_y,
                self.section.i_x,
                4.00,
                4.00,
                2.00,
                k11_x,
                k33_x,
                k44_x,
                self.geomtransf.uid,
            ]

        if self.n_y is not None and self.n_x is None:
            return [
                "ModElasticBeam3d",
                self.uid,
                self.nodes[0].uid,
                self.nodes[1].uid,
                self.section.area,
                self.section.e_mod,
                self.section.g_mod,
                self.section.j_mod,
                self.section.i_y,
                self.section.i_x,
                k11_y,
                k33_y,
                k44_y,
                4.00,
                4.00,
                2.00,
                self.geomtransf.uid,
            ]

        if self.n_x is not None and self.n_y is not None:
            return [
                "ModElasticBeam3d",
                self.uid,
                self.nodes[0].uid,
                self.nodes[1].uid,
                self.section.area,
                self.section.e_mod,
                self.section.g_mod,
                self.section.j_mod,
                self.section.i_y,
                self.section.i_x,
                k11_y,
                k33_y,
                k44_y,
                k11_x,
                k33_x,
                k44_x,
                self.geomtransf.uid,
            ]

        return [
            "elasticBeamColumn",
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.section.area,
            self.section.e_mod,
            self.section.g_mod,
            self.section.j_mod,
            self.section.i_y,
            self.section.i_x,
            self.geomtransf.uid,
        ]

    def clear_length(self):
        """
        Returns the clear length of the element (without the rigid
        offsets)

        """

        p_i = np.array(self.nodes[0].coords) + self.geomtransf.offset_i
        p_j = np.array(self.nodes[1].coords) + self.geomtransf.offset_j
        return np.linalg.norm(p_i - p_j)

    def __repr__(self):
        res = ""
        res += "elasticBeamColumn element object\n"
        res += f"uid: {self.uid}\n"
        res += f"node_i.uid: {self.nodes[0].uid}\n"
        res += f"node_j.uid: {self.nodes[1].uid}\n"
        res += f"node_i.coords: {self.nodes[0].coords}\n"
        res += f"node_j.coords: {self.nodes[1].coords}\n"
        res += f"offset_i: {self.geomtransf.offset_i}\n"
        res += f"offset_j: {self.geomtransf.offset_j}\n"
        res += f"x_axis: {self.geomtransf.x_axis}\n"
        res += f"y_axis: {self.geomtransf.y_axis}\n"
        res += f"z_axis: {self.geomtransf.z_axis}\n"
        res += f"section.name: {self.section.name}\n"
        return res


@dataclass
class BeamIntegration:
    """
    OpenSees beamIntegration parent class.
    https://openseespydoc.readthedocs.io/en/latest/src/beamIntegration.html?highlight=beamintegration

    """

    uid: int
    parent_section: Union[ElasticSection, FiberSection] = field(repr=False)


@dataclass
class Lobatto(BeamIntegration):
    """
    OpenSees Lobatto beam integration.
    https://openseespydoc.readthedocs.io/en/latest/src/Lobatto.html

    """

    n_p: int

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return ["Lobatto", self.uid, self.parent_section.uid, self.n_p]


@dataclass
class DispBeamColumn(Element):
    """
    OpenSees dispBeamColumn element
    https://openseespydoc.readthedocs.io/en/latest/src/ForceBeamColumn.html

    """

    section: FiberSection
    geomtransf: GeomTransf
    integration: BeamIntegration

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees

        """

        return [
            "dispBeamColumn",
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.geomtransf.uid,
            self.integration.uid,
            # '-iter',  # can change it to forceBeamColumn here for testing
            # 50,
            # 1e-1
        ]

    def clear_length(self):
        """
        Returns the clear length of the element (without the rigid
        offsets)

        """

        p_i = np.array(self.nodes[0].coords) + self.geomtransf.offset_i
        p_j = np.array(self.nodes[1].coords) + self.geomtransf.offset_j
        return np.linalg.norm(p_i - p_j)

    def __repr__(self):
        res = ""
        res += "dispBeamColumn element object\n"
        res += f"uid: {self.uid}\n"
        res += f"node_i.uid: {self.nodes[0].uid}\n"
        res += f"node_j.uid: {self.nodes[1].uid}\n"
        res += f"node_i.coords: {self.nodes[0].coords}\n"
        res += f"node_j.coords: {self.nodes[1].coords}\n"
        res += f"offset_i: {self.geomtransf.offset_i}\n"
        res += f"offset_j: {self.geomtransf.offset_j}\n"
        res += f"x_axis: {self.geomtransf.x_axis}\n"
        res += f"y_axis: {self.geomtransf.y_axis}\n"
        res += f"z_axis: {self.geomtransf.z_axis}\n"
        res += f"section.name: {self.section.name}\n"
        return res
